import sys
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments, loadUI
    import MJOLNIRGui.src.main.python._tools as _GUItools
except ImportError:
    from _tools import ProgressBarDecoratorArguments, loadUI
    import _tools as _GUItools
from os import path
from PyQt5 import QtWidgets,uic,QtGui,QtCore
import numpy as np
from MJOLNIR._tools import calculateMolarMass
import pyperclip

# Handles all functionality related to the NormalizationManager. 

def sampleFormRadioButtonChange(self,state,formLE,massSB): # Function for radio button on form. If set enable LineEdit of form, disable mass SB or reverse
    formLE.setDisabled(not state)
    massSB.setDisabled(state)
    # If state is switched to chemical formula use string in field
    if state: 
        checkValidSampleNormalization(formLE,None,massSB)


# def updateToDataSet(self,formName,event):
#     ds = self.guiWindow.DataSetModel.getCurrentDataSet()
#     attribute = getattr(self.guiWindow.ui,'NormalizationManager_'+formName)
#     if not ds is None: # Only when there is a data set
#         if formName.find('_lineEdit')>0:
#             value=attribute.text().strip()
#             if value == '': # Use standard text instead
#                 value = attribute.placeholderText().strip()
#             print(value)
#         elif formName.find('_spinBox')>0:
#             value=attribute.value()
#         elif formName.find('_checkBox')>0:
#             value=attribute.isChecked()
#         else:
#             value = None
#     ds.currentNormalizationSettings[formName] = value
    
#     attribute.focusOutEvent_old(event)

def checkValidSampleNormalization(self,event,massSB):
    string = self.text()
    try:
        mass = calculateMolarMass(string)
        self.setStyleSheet("color: black;")
    except AttributeError as e:
        self.setStyleSheet("color: red;")
        _GUItools.dialog('When parsing "{}" following error occured:\n'.format(string)+str(e))
        if not event is None:
            self.focusOutEvent_old2(event)
        return None

    if len(string)!=0 and mass==0: # A string was provided but no mass found, typically error in input
        self.setStyleSheet("color: red;")
        _GUItools.dialog('When parsing "{}" no mass was found. Please format the chemical formula correctly, e.g. MnF2'.format(string))
    elif len(string)==0:
        mass = calculateMolarMass(self.placeholderText())
        massSB.setValue(mass)
    else:
        massSB.setValue(mass)
    if not event is None:
        self.focusOutEvent_old2(event)

NormalizationManagerBase, NormalizationManagerForm = loadUI('NormalizationManager.ui')

# # All of this connects the buttons and their functions to the main window.

class NormalizationManager(NormalizationManagerBase, NormalizationManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(NormalizationManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/ruler.png')))
        self.initNormalizationManager()
        
    def initNormalizationManager(self):    
        self.sampleGroupBox = self.NormalizationManager_sampleGroupBox_checkBox
        self.nSampleGroupBox = self.NormalizationManager_normalizationSampleGroupBox_checkBox
        self.normalizationButton= self.NormalizationManager_applyNormalization_btn
        self.setup()
        

    def setup(self):
        # # Connect update of value in field with corresponding value in dataset
        # # As this cannot be done in a loop, everything is written out

        # self.NormalizationManager_sampleFormula_lineEdit.focusOutEvent_old = self.NormalizationManager_sampleFormula_lineEdit.focusOutEvent
        # self.NormalizationManager_sampleFormula_lineEdit.focusOutEvent = lambda event: updateToDataSet(self,'sampleFormula_lineEdit',event)

        # self.NormalizationManager_sampleMolarMass_spinBox.focusOutEvent_old = self.NormalizationManager_sampleMolarMass_spinBox.focusOutEvent
        # self.NormalizationManager_sampleMolarMass_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'sampleMolarMass_spinBox',event)

        # self.NormalizationManager_sampleMass_spinBox.focusOutEvent_old = self.NormalizationManager_sampleMass_spinBox.focusOutEvent
        # self.NormalizationManager_sampleMass_spinBox.focusOutEvent= lambda event: updateToDataSet(self,'sampleMass_spinBox',event)

        # self.NormalizationManager_sampleUnitsPerCell_spinBox.focusOutEvent_old = self.NormalizationManager_sampleUnitsPerCell_spinBox.focusOutEvent
        # self.NormalizationManager_sampleUnitsPerCell_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'sampleUnitsPerCell_spinBox',event)

        # self.NormalizationManager_sampleGFactor_spinBox.focusOutEvent_old = self.NormalizationManager_sampleGFactor_spinBox.focusOutEvent
        # self.NormalizationManager_sampleGFactor_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'sampleGFactor_spinBox',event)


        # self.NormalizationManager_normalizationSampleMolarMass_spinBox.focusOutEvent_old = self.NormalizationManager_normalizationSampleMolarMass_spinBox.focusOutEvent
        # self.NormalizationManager_normalizationSampleMolarMass_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'normalizationSampleMolarMass_spinBox',event)

        # self.NormalizationManager_normalizationSampleMass_spinBox.focusOutEvent_old = self.NormalizationManager_normalizationSampleMass_spinBox.focusOutEvent
        # self.NormalizationManager_normalizationSampleMass_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'normalizationSampleMass_spinBox',event)

        # self.NormalizationManager_normalizationSampleUnitsPerCell_spinBox.focusOutEvent_old = self.NormalizationManager_normalizationSampleUnitsPerCell_spinBox.focusOutEvent
        # self.NormalizationManager_normalizationSampleUnitsPerCell_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'normalizationSampleUnitsPerCell_spinBox',event)

        # self.NormalizationManager_normalizationSampleGFactor_spinBox.focusOutEvent_old = self.NormalizationManager_normalizationSampleGFactor_spinBox.focusOutEvent
        # self.NormalizationManager_normalizationSampleGFactor_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'normalizationSampleGFactor_spinBox',event)

        # self.NormalizationManager_normalizationMonitor_spinBox.focusOutEvent_old = self.NormalizationManager_normalizationMonitor_spinBox.focusOutEvent
        # self.NormalizationManager_normalizationMonitor_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'normalizationMonitor_spinBox',event)

        # self.NormalizationManager_normalizationSampleSigmaInc_spinBox.focusOutEvent_old = self.NormalizationManager_normalizationSampleSigmaInc_spinBox.focusOutEvent
        # self.NormalizationManager_normalizationSampleSigmaInc_spinBox.focusOutEvent = lambda event: updateToDataSet(self,'normalizationSampleSigmaInc_spinBox',event)

        # self.NormalizationManager_normalizationSampleFormula_lineEdit.focusOutEvent_old = self.NormalizationManager_normalizationSampleFormula_lineEdit.focusOutEvent
        # self.NormalizationManager_normalizationSampleFormula_lineEdit.focusOutEvent = lambda event: updateToDataSet(self,'normalizationSampleFormula_lineEdit',event)
        
        
        self.sampleGroupBox.clicked.connect(lambda event: self.checkNormalizable(event))
        self.nSampleGroupBox.clicked.connect(lambda event: self.checkNormalizable(event))



        #self.normalizationCheckBox.focusOutEvent_old = lambda event: None
        #self.normalizationCheckBox.clicked.connect(lambda event: updateToDataSet(self,'applyNormalization_checkBox',event))
        



        lostFocus = QtGui.QFocusEvent(QtGui.QEnterEvent.FocusOut) # Event used for return pressed functionality
        ## Shorten naming
        sampleFormLE = self.NormalizationManager_sampleFormula_lineEdit
        sampleMassSB = self.NormalizationManager_sampleMolarMass_spinBox
        sampleFormRB = self.NormalizationManager_sampleFormula_radioButton
        # Rename old focusOutEvent to be used in checkValidSample function
        sampleFormLE.focusOutEvent_old2 = sampleFormLE.focusOutEvent
        # Connect the check validater when focus is lost or enter is pressed
        sampleFormLE.focusOutEvent= lambda event:checkValidSampleNormalization(sampleFormLE,event,sampleMassSB)
        sampleFormLE.returnPressed.connect(lambda: checkValidSampleNormalization(sampleFormLE,lostFocus,sampleMassSB))
        

        # Repeat for normalization fields/buttons
        nSsampleFormLE = self.NormalizationManager_normalizationSampleFormula_lineEdit
        nSampleMassSB = self.NormalizationManager_normalizationSampleMolarMass_spinBox
        nSampleFormRB = self.NormalizationManager_normalizationSampleFormula_radioButton

        nSsampleFormLE.focusOutEvent_old2 = nSsampleFormLE.focusOutEvent
        nSsampleFormLE.focusOutEvent = lambda event:checkValidSampleNormalization(nSsampleFormLE,event,nSampleMassSB)
        nSsampleFormLE.returnPressed.connect(lambda: checkValidSampleNormalization(nSsampleFormLE,lostFocus,nSampleMassSB))
        
        
        # Change enabled fields when radio buttons are activated
        sampleFormRB.toggled.connect(lambda state:sampleFormRadioButtonChange(sampleFormRB,state,formLE=sampleFormLE,massSB=sampleMassSB))
        nSampleFormRB.toggled.connect(lambda state:sampleFormRadioButtonChange(nSampleFormRB,state,formLE=nSsampleFormLE,massSB=nSampleMassSB))

        normalizationButton = self.NormalizationManager_applyNormalization_btn

        normalizationButton.clicked.connect(self.saveNormalization)


        # Enable normalizationButton only when sampleGroupBox is checked ###when either sampleGroupBox or nSampleGroupBox is activated
        #self.sampleGroupBox.clicked.connect(lambda event: groupBoxCheckChanged(self.sampleGroupBox,event,normalizationButton,[self.sampleGroupBox,self.nSampleGroupBox]))
        #self.nSampleGroupBox.clicked.connect(lambda event: groupBoxCheckChanged(self.nSampleGroupBox,event,normalizationButton,[self.sampleGroupBox,self.nSampleGroupBox]))


        # Without a dataset active, set the check boxes disabled
        self.sampleGroupBox.setDisabled(False)
        self.nSampleGroupBox.setDisabled(False)

        #normalizationButton.updateChecked = (lambda state: autoEnable(normalizationButton,state))
        #normalizationButton.toggled.connect(normalizationButton.updateChecked)

        
    def getSampleFormNames(self,sample=True):

        # All line edit forms, spinboxes
        lEFormNames = ['sampleFormula_lineEdit']
        sBFormNames = ['sampleMolarMass_spinBox','sampleMass_spinBox','sampleUnitsPerCell_spinBox','sampleGFactor_spinBox']
        cBFormNames = []
        if sample is False: # Change names to fit to normalization sample
            lEFormNames = [x.replace('sample','normalizationSample') for x in lEFormNames]
            sBFormNames = [x.replace('sample','normalizationSample') for x in sBFormNames]

            sBFormNames.append('normalizationMonitor_spinBox')
            sBFormNames.append('normalizationSampleSigmaInc_spinBox')
        return lEFormNames,sBFormNames,cBFormNames

    def getSampleDict(self,sample=True): # If sample is false, get info from normalization sample
        lEFormNames,sBFormNames, cBFormNames = self.getSampleFormNames(sample)
        sampleDict = {}
        # add values to sampleDict
        for lE in lEFormNames:
            value = getattr(self,'NormalizationManager_'+lE).text().strip()
            if value is None or value=='':
                value = getattr(self,'NormalizationManager_'+lE).placeholderText().strip()
            sampleDict[lE] = value
        
        for sB in sBFormNames:
            sampleDict[sB] = getattr(self,'NormalizationManager_'+sB).value()

        for cB in cBFormNames:
            sampleDict[cB] = getattr(self,'NormalizationManager_'+cB).isChecked()

        #print(sampleDict)
        return sampleDict

    def getSampleInputs(self): # Get all inputs from the fields under sample
        
        
        sampleDict = {}
        # Check if renormalization is applied at all
        #apply = self.normalizationCheckBox.isChecked()
        #sampleDict['applyNormalization_checkBox'] = apply
        sampleDict['sampleGroupBox_checkBox'] = self.sampleGroupBox.isChecked()
        sampleDict['normalizationSampleGroupBox_checkBox'] = self.nSampleGroupBox.isChecked()
        sampleDict.update(self.getSampleDict())
        sampleDict.update(self.getSampleDict(sample=False))

        return sampleDict

    def getActiveNormalization(self):
        sampleDict = {}
        sampleDict['sampleGroupBox_checkBox'] = self.sampleGroupBox.isChecked()
        sampleDict['normalizationSampleGroupBox_checkBox'] = self.nSampleGroupBox.isChecked()


    def saveNormalization(self):
        #print('Hmm I should save this....')
        inputs = self.getSampleInputs()

        returnString = []
        arguments = []

        if inputs['sampleGroupBox_checkBox']: # The sample check box has been choosen
            returnString.append('# Write values for sample normalization\n')
            # Figure out if chemical formula or molar mass is to be used
            if self.NormalizationManager_sampleFormula_radioButton.isChecked(): # if true chemcial formula is to be used
                sampleFormula = inputs['sampleFormula_lineEdit']
                sampleMolarMass = None
                returnString.append('sampleChemicalFormula = "'+sampleFormula+'"')
                arguments.append('sampleChemicalFormula = sampleFormula')
            else:
                sampleFormula = None
                sampleMolarMass = inputs['sampleMolarMass_spinBox']
                returnString.append('sampleMolarMass = '+str(sampleMolarMass)+' # g/mol')
                arguments.append('sampleMolarMass = sampleMolarMass')
            sampleMass = inputs['sampleMass_spinBox']
            sampleUnitsPerCell = inputs['sampleUnitsPerCell_spinBox']
            sampleGFactor = inputs['sampleGFactor_spinBox']

            returnString.append('sampleMass = '+str(sampleMass)+' # g')
            arguments.append('sampleMass = sampleMass')
            returnString.append('sampleUnitsPerCell = '+str(sampleUnitsPerCell))
            arguments.append('formulaUnitsPerUnitCell = sampleUnitsPerCell')
            returnString.append('sampleGFactor = '+str(sampleGFactor))
            arguments.append('sampleGFactor = sampleGFactor')

            returnString.append('')

        if inputs['normalizationSampleGroupBox_checkBox']: # 
            

            if self.NormalizationManager_normalizationSampleFormula_radioButton.isChecked(): # if true chemcial formula is to be used
                nSampleFormula = inputs['normalizationSampleFormula_lineEdit']
                nSampleMolarMass = None
                returnString.append('normalizationSampleFormula = "'+nSampleFormula+'"')
                arguments.append('vanadiumChemicalFormula = normalizationSampleFormula')
            else:
                nSampleFormula = None
                nSampleMolarMass = inputs['normalizationSampleMolarMass_spinBox']
                returnString.append('normalizationSampleMolarMass = '+str(nSampleMolarMass)+' # g/mol')
                arguments.append('vanadiumMolarMass = normalizationSampleMolarMass')

            nSampleMass = inputs['normalizationSampleMass_spinBox']
            nSampleUnitsPerCell = inputs['sampleUnitsPerCell_spinBox']
            nSampleGFactor = inputs['sampleGFactor_spinBox']
            nSampleMonitor = inputs['normalizationMonitor_spinBox']
            nSampleSigma = inputs['normalizationSampleSigmaInc_spinBox']

            returnString.append('normalizationSampleMass = '+str(nSampleMass)+' # g')
            arguments.append('vanadiumMass = normalizationSampleMass')
            returnString.append('normalizationSampleUnitsPerCell = '+str(nSampleUnitsPerCell))
            arguments.append('vanadiumUnitsPerUnitCell = normalizationSampleUnitsPerCell')
            returnString.append('normalizationSampleGFactor = '+str(nSampleGFactor))
            arguments.append('vanadiumGFactor = normalizationSampleGFactor')
            returnString.append('normalizationSampleMonitor = '+str(nSampleMonitor)+' # Neutron counts in monitor/scan step')
            arguments.append('vanadiumMonitor = normalizationSampleMonitor')
            returnString.append('sigma = '+str(nSampleSigma)+' # Incoherent scattering in Barns')
            arguments.append('vanadiumSigmaIncoherent = sigma')
            returnString.append('')

        else:
            arguments.append('correctVanadium = False')

        args = ''
        arguments.reverse()
        while len(arguments)>0:
           line = ''
           while(len(line)<65) and len(arguments)>0:
               line+=arguments.pop()+', '
           args+=line+'\n'
            
        # Add only to -3 as the last two characters will be ', \n'
        returnString.append('ds.absoluteNormalize('+args[:-3]+')')
        #print(arguments)

        pyperclip.copy('\n'.join(returnString))

        self.normalizationButton.setStyleSheet("color: green;")
        self.timer=QtCore.QTimer()
        self.timer.timeout.connect(self.resetApplyButton)
        self.timer.start(1000)
            
            

    def resetApplyButton(self):
        self.normalizationButton.setStyleSheet("color: black;")

    #

    # def updateCurrentDataSetIndex(self):
    #     ds = self.guiWindow.DataSetModel.getCurrentDataSet()
    #     if ds is None:
    #         self.sampleGroupBox.setDisabled(True)
    #         self.nSampleGroupBox.setDisabled(True)
    #     elif len(ds.convertedFiles)==0:
    #         self.sampleGroupBox.setDisabled(True)
    #         self.nSampleGroupBox.setDisabled(True)
    #     else:
    #         self.sampleGroupBox.setDisabled(False)
    #         self.nSampleGroupBox.setDisabled(False)

    #         self.sampleGroupBox.setChecked(bool(ds.currentNormalizationSettings['sampleGroupBox_checkBox']))
    #         self.nSampleGroupBox.setChecked(bool(ds.currentNormalizationSettings['normalizationSampleGroupBox_checkBox']))
            
    #         for sample in [True,False]:
    #             lEFormNames, sBFormNames, cBFormNames = self.getSampleFormNames(sample=sample)
    #             for lE in lEFormNames:
    #                 getattr(self.guiWindow.ui,'NormalizationManager_'+lE).setText(str(ds.currentNormalizationSettings[lE]))
    #             for sB in sBFormNames:
    #                 getattr(self.guiWindow.ui,'NormalizationManager_'+sB).setValue(float(ds.currentNormalizationSettings[sB]))
    #             for cB in cBFormNames:
    #                 getattr(self.guiWindow.ui,'NormalizationManager_'+cB).setChecked(bool(ds.currentNormalizationSettings[cB]))
    
    #     self.checkNormalizable()

    def checkNormalizable(self,event=None):
    #     ds = self.guiWindow.DataSetModel.getCurrentDataSet()
    #     if ds is None:
    #         self.normalizationButton.setText('Apply Normalization')
    #         self.normalizationButton.setDisabled(True)
    #         return None
        
        
    #     ds.currentNormalizationSettings['sampleGroupBox_checkBox'] = self.sampleGroupBox.isChecked()
    #     ds.currentNormalizationSettings['normalizationSampleGroupBox_checkBox'] = self.nSampleGroupBox.isChecked()

    #     normalizable = ds.isNormalizable()
        checkBoxes = [self.sampleGroupBox,self.nSampleGroupBox]
        allowed = np.any([cB.isChecked() for cB in checkBoxes])# and self.guiWindow.stateMachine.getCurrentState().name == 'Converted'
        if allowed == 0:
            self.normalizationButton.setText('Copy to Clipboard')
            self.normalizationButton.setDisabled(True)
            return None
        else:
    #     # If this point is reached, the data set can be normalized or reverted
    # 
    #     if normalizable == -1: # already normalized and has same settings as before
    #         self.normalizationButton.setText('Revert Normalization')
    #         self.normalizationButton.setDisabled(False)
    #     elif normalizable == 1: 
            self.normalizationButton.setText('Copy to Clipboard')
            self.normalizationButton.setDisabled(False)
        

    # def applyNormalization(self):
    #     self.guiWindow.DataSetModel.applyNormalization()
        
    #     self.checkNormalizable()



from fbs.cmdline import command
from os.path import dirname
from fbs.builtin_commands import freeze,clean
import fbs.cmdline
from shutil import copyfile

import os,sys
import MJOLNIR

@command
def customFreeze(debug=False):
    print('Cleaning up the previous mess...')
    clean()
    print('Freezing application with debug = {}'.format(debug))
    freeze(debug=debug)
    print('Generating needed folder for MJOLNIR normalizations')

    operatingSystem = sys.platform

    if operatingSystem == 'darwin':
        newFolder = os.path.join('target','MJOLNIRGui.app','Contents','MacOS','MJOLNIR')
    else:
        newFolder = os.path.join('target','MJOLNIRGui','MJOLNIR')
    if not os.path.isdir(newFolder):
        os.mkdir(newFolder)
    for f in [MJOLNIR.__flatConeNormalization__,MJOLNIR.__bambusNormalization__,MJOLNIR.__multiFLEXXNormalization__,MJOLNIR.__CAMEANormalization__]:
        print(f)
        copyfile(f, os.path.join(newFolder,os.path.split(f)[1]))
    
    
    if operatingSystem == sys.platform == 'win32':
        location = os.path.join('target','MJOLNIRGui','MJOLNIRGui.exe')
    else:
        location = os.path.join('target','MJOLNIRGui','MJOLNIRGui')
    print("Done. You can now run `{}`. If that\ndoesn't work, see https://build-system.fman.io/troubleshooting.".format(location))



if __name__ == '__main__':
    project_dir = dirname(__file__)
    print(project_dir)
    fbs.cmdline.main(project_dir)


import pickle as pickle
import os,traceback
import functools
from PyQt5 import QtCore,QtWidgets

from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QVBoxLayout, QCheckBox, QLabel,QApplication
from PyQt5 import uic
import platform
from os import path

def ProgressBarDecoratorArguments(runningText='Running',completedText='Completed',failedText='Failed'):

    def ProgressBarDecorator(func):
        @functools.wraps(func)
        def newFunc(self,*args,**kwargs):
            self.setProgressBarValue(0)
            self.setProgressBarLabelText(runningText)
            if len(args) == 1:
                args = ()
            else:
                args = args[1:]
            self.update()
            try:
                returnval = func(self,*args,**kwargs)
            except Exception:
                err = traceback.format_exc()
                state = self.stateMachine.currentState.name
                errText = 'MJOLNIRGui encountered an error with the following error message:\n\n{}\nMJOLNIRGui was in:{}\n\n'.format(str(err),state)+\
                    'If this is a recurring error, that you believe should be fixed, please  feel free to copy the message and '+\
                    'send it in an email to "MJOLNIRPackage@gmail.com".'
                
                dialog(errText)
                returnval = False

            self.setProgressBarMaximum(100)
            if returnval is not None:
                if returnval is False:
                    self.setProgressBarValue(0)
                    self.setProgressBarLabelText(failedText)
                    self.resetProgressBarTimed()
                    return returnval
        
            self.setProgressBarValue(100)
            self.setProgressBarLabelText(completedText)
            self.resetProgressBarTimed()
            return returnval
        return newFunc
    return ProgressBarDecorator


def loadSetting(settingsFile,string): # pragma: no cover
    if Exists(settingsFile):
        with open(settingsFile,"rb") as pickle_in:
            loaded_dict = pickle.load(pickle_in)
        if string in loaded_dict:
            returnValue = loaded_dict[string]
        else:
            returnValue = None
        return returnValue
    else:
        return None

def updateSetting(settingsFile, name,value):# pragma: no cover
    if Exists(settingsFile):
        with open(settingsFile,"rb") as pickle_in:
            loaded_dict = pickle.load(pickle_in)
    else:
        loaded_dict = {}
    loaded_dict[name]=value
    with open(settingsFile,"wb") as pickle_out:
        pickle.dump(loaded_dict, pickle_out,protocol=pickle.HIGHEST_PROTOCOL)

def Exists(settingsFile):# pragma: no cover
    return os.path.isfile(settingsFile)

def dialog(text):
    dialog = QtWidgets.QMessageBox()
    dialog.setStyleSheet("messagebox-text-interaction-flags: 5")
    dialog.setIcon(QtWidgets.QMessageBox.Critical)
    dialog.setText(text)
    dialog.addButton(QtWidgets.QMessageBox.Ok)
    dialog.exec() 

def CenterWidgets(widget, host = None):
    if (host is None):
        host = widget.parentWidget()

    if (not host is None):
        hostRect = host.geometry()
        widget.move(hostRect.center() - widget.rect().center())
    
    else:
        screenGeometry = QApplication.desktop().screenGeometry()
        x = (screenGeometry.width() - widget.width()) / 2
        y = (screenGeometry.height() - widget.height()) / 2
        widget.move(x, y)
    

def loadUI(fileName):
    # print(platform.system().lower(),fileName)
    if platform.system().lower() == 'darwin':
        folder = path.abspath(path.join(path.dirname(__file__),'..','Resources','Views'))
    else: 
        folder = path.join(path.dirname(__file__),'..','..','resources','base','Views')


    try:
    # needed before freezing app
        base,form = uic.loadUiType(path.join(path.dirname(__file__),fileName))
    
    except FileNotFoundError:
        try:
            # needed when running app local through fbs
            base,form = uic.loadUiType(path.abspath(path.join(folder,fileName)))
            
        except FileNotFoundError:
            # needed when running app after pip install
            
            # except FileNotFoundError:
            #     base,form = uic.loadUiType(path.join(folder,fileName))
            try:
                base,form = uic.loadUiType(path.abspath(path.join(path.dirname(__file__),'..','resources','base','Views',fileName)))
                
            except FileNotFoundError:
                base,form = uic.loadUiType(path.abspath(path.join(path.dirname(__file__),'Views',fileName)))
    return base,form
        

class log(list):
    def __init__(self,textBrowser,button,maxLines = 80):
        self.maxLines = maxLines
        self.textBrowser = textBrowser
        self.button = button
        
        self.button.setDisabled(False)
        
    def clear(self):
        while len(self)>0:
            self.pop()

        self.updateTextBrowser()

    def trim(self):
        while len(self)>self.maxLines:
            self.pop()

    def updateTextBrowser(self):
        self.textBrowser.setText('\n'.join(self))
        if len(self) == 0:
            self.button.setDisabled(True)
        else:
            self.button.setDisabled(False)

    def __setitem__(self, key, value):
        super(log, self).__setitem__(key, value)
        self.updateTextBrowser()

    def __delitem__(self, value):
        super(log, self).__delitem__(value)
        self.updateTextBrowser()

    def __add__(self, value):
        super(log, self).__add__(value)
        self.trim()
        self.updateTextBrowser()
        

    def __iadd__(self, value):
        super(log, self).__iadd__(value)
        self.trim()
        self.updateTextBrowser()

    def append(self, value):
        super(log, self).insert(0,value)
        self.trim()
        self.updateTextBrowser()

    def remove(self, value):
        super(log, self).remove(value)
        self.updateTextBrowser()
        



from PyQt5 import QtCore
from MJOLNIR.Data import DataSet,DataFile,Mask
import copy
import numpy as np
#from collections import defaultdict

class GuiDataSet(DataSet.DataSet):
    def __init__(self,dataFiles=None,name='No Name',dataSet=None,background=None,**kwargs):
        if not dataSet is None:
            self.updateProperty(dataSet.__dict__)
        else:
            super(GuiDataSet,self).__init__(dataFiles=dataFiles,**kwargs)
        self.name = name
        
        # If generated by subtraction, save the background data file locations here
        self.background = background
        
        for idx,df in enumerate(self):
            df.idx = idx

        self._maskingObject = None # Holder for masking
        
    def setData(self,column,value):
        if column == 0: self.name = value
        
    def convertDataFile(self,dataFiles=None,binning=None,guiWindow=None,setProgressBarMaximum=True,progressUpdate=1,printFunction=None):
        """
        
        Args:

            - dataFiles (list): Files to be update. If None, use self.dataFiles (default None)

            - guiWindow (window): Current guiwindow used to access progressbar

            - setProgressBarMaximum (bool): if True, overwrite curernt progressbar max (default True)

            - progressUpdate (float): Amount to update progress bar (default 1)

            - printFunction (function): Function called with text if any is generated during conversin (Default None, -> warning)
        """
        
        dataFiles = list(self)
        if not guiWindow is None and setProgressBarMaximum:
            guiWindow.setProgressBarMaximum(len(dataFiles)+1)

        convertedFiles = []
        for _,rawfile in enumerate(dataFiles):
            convFile = rawfile.convert(printFunction=printFunction,binning=binning)
                
            convertedFiles.append(convFile)
            if not guiWindow is None:
                guiWindow.addProgressBarValue(progressUpdate)
            
        self._convertedFiles = []
        if len(convertedFiles)!=0:
            self.convertedFiles = convertedFiles    
        
        if len(self)!=0:
            self._getData()
        #if not guiWindow is None:
        #    guiWindow.setProgressBarValue(len(dataFiles))


    def flags(self):
        return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsSelectable

    def insert(self,target_row,item,**kwargs):
        if len(self.convertedFiles)>0:
            self.dataFiles.insert(target_row,item.original_file,**kwargs)
            return self.convertedFiles.insert(target_row,item,**kwargs)
        else:
            return self.dataFiles.insert(target_row,item,**kwargs)

    def pop(self,position):
        if len(self.convertedFiles)>0:
            self.dataFiles.pop(position)
            return self.convertedFiles.pop(position)
        else:
            return self.dataFiles.pop(position)

    def append(self,files):
        convert = False # If added files should be converted
        if len(self)>0:
            if self[0].type == 'nxs':
                convert = self[0].binning
        super(GuiDataSet,self).append(files)
        for idx,df in enumerate(self):
            df.idx = idx

        if convert:
            self.convertDataFile(dataFiles=files,binning=convert)

    def updateProperty(self,dictionary):
        if isinstance(dictionary,dict):
            for key in dictionary.keys():
                self.__setattr__(key,copy.deepcopy(dictionary[key]))

    # def undoAbsolutNormalize(self):
    #     if len(self.convertedFiles)>0: # If there are converted files present, undo normalization
    #         super(GuiDataSet,self).undoAbsolutNormalize()

    #     self.normalizationSettings = {}
    #     print(self.normalizationSettings,self.currentNormalizationSettings)
        
    # def isNormalizable(self): # checks if the current data set can be normalized
    #     # If the current normalization settings are non-zero and equal to previously used values, return -1 to signify an allowance for revert of normalization
    #     if self.currentNormalizationSettings != {} and self.currentNormalizationSettings==self.normalizationSettings:
    #         return -1
    #     if self.currentNormalizationSettings == {}:
    #         return 0
    #     else:
    #         return 1

    # def absolutNormalize(self):
    #     if self.normalizationSettings == self.currentNormalizationSettings:
    #         ## Instead of normalize revert normalization
    #         self.undoAbsolutNormalize()
    #         normalizationParams = self.parent().sampleManager.getSampleInputs()
    #         self.currentNormalizationSettings.update(normalizationParams)
    #         return

    #     self.normalizationSettings.update(self.currentNormalizationSettings)

    #     raise NotImplementedError('Currently not implemented')
    #     sampleMass = self.currentNormalizationSettings['sampleMass_spinBox']
    #     units = self.sampleUnitsPerCell_spinBox
    #     gfactor = self.sampleGFactor_spinBox
    #     chemicalFormula = self.sampleFormula_lineEdit
        
    #     print("Normalized DataSet",sampleMass,units,gfactor,chemicalFormula)

    #     if hasattr(self,'SampleManager_normalizationSampleGroupBox_checkBox'):
    #         if self.SampleManager_normalizationSampleGroupBox_checkBox:
    #             #nMolarMass = self.normalizationSampleMolarMass_spinBox
    #             nChemicalFormula = self.normalizationSampleFormula_lineEdit
    #             nSampleMass = self.normalizationSampleMass_spinBox
    #             nUnits = self.normalizationSampleUnitsPerCell_spinBox
    #             nGfactor = self.normalizationSampleGFactor_spinBox
    #             nMonitor = self.normalizationMonitor_spinBox
    #             nSigma = self.normalizationSampleSigmaInc_spinBox
    #             super(GuiDataSet,self).absolutNormalize(sampleMass=sampleMass,sampleChemicalFormula=chemicalFormula,
    #             formulaUnitsPerUnitCell=units,sampleGFactor=gfactor,correctVanadium=True,vanadiumMass=nSampleMass,vanadiumChemicalFormula=nChemicalFormula,
    #             vanadiumMonitor=nMonitor,vanadiumSigmaIncoherent=nSigma,vanadiumGFactor=nGfactor,vanadiumUnitsPerUnitCell=nUnits)

    #     else:
    #         super(GuiDataSet,self).absolutNormalize(sampleMass=sampleMass,sampleChemicalFormula=chemicalFormula,
    #             formulaUnitsPerUnitCell=units,sampleGFactor=gfactor,correctVanadium=False)
                
class GuiDataFile(DataFile.DataFile):
    def __init__(self,fileLocation, **kwargs):
        self._idx = 0
        super(GuiDataFile,self).__init__(fileLocation=fileLocation,**kwargs)
        binning = 1
        calibrationIndex = list(self.possibleBinnings).index(binning) # Only binning 1 is used for raw plotting
        
        if self.instrument == 'CAMEA':
            EPrDetector = 8 
            detectors = 104
        elif self.type == 'MultiFLEXX':
            EPrDetector = 5
            detectors = 31
        elif self.type == 'FlatCone':
            EPrDetector = 1
            detectors = 31
        elif self.type == 'Babus':
            EPrDetector = 1
            detectors = 20
        else:

            totalDetectors = np.array(self.instrumentCalibrations[calibrationIndex][0].shape[:-1])
            if np.mod(totalDetectors,31)==0: # either MultiFLEXX or FlatCone
                EPrDetector = int(totalDetectors/31)
                detectors = 31
            else: # CAMEA
                EPrDetector = 8 
                detectors = 104
        
        self.maxDetectorSelection = detectors
        self.maxAnalyzerSelection = EPrDetector
        self.detectorSelectionOriginal = self.detectorSelection
        self.analyzerSelectionOriginal = self.analyzerSelection

    @property
    def idx(self):
        return self._idx

    @idx.setter
    def idx(self,value):
        if self.type == 'nxs':
            self.original_file.idx = value
        else:
            self._idx = value
    
    @idx.getter
    def idx(self):
        if self.type == 'nxs':
            return self.original_file.idx
        else:
            return self._idx

    def setData(self,column,value):
        if column == 0: self.name = value

    def flags(self):
        return QtCore.Qt.ItemIsEditable

    def convert(self, printFunction,binning=None):
        if self.type == 'nxs':
            df = self.original_file
        else:
            df = self
        if binning is None:
            binning = df.binning
        convertedFile = GuiDataFile(super(GuiDataFile, df).convert(binning = binning,printFunction=printFunction))
        
        return convertedFile

class Gui1DCutObject(object):
    def __init__(self,name,parameters,pdData,bins):
        self.name = name
        self.parameters = parameters
        self._ufit = None
        self.pdData = pdData
        self.bins = bins
    
    def plot(self,*args,**kwargs):
        # redo to capitalize only first letter of the saved method (cut1D -> plotCut1D)
        if self.parameters['method'].find('plot')<0:
            plotName = 'plot'+self.parameters['method'][0].capitalize()+self.parameters['method'][1:]
            self.parameters['method'] = plotName
        else:
            plotName = self.parameters['method']
        plotFunction = getattr(self.parameters['dataset'],plotName)
        if self.parameters['method'].find('ut1DE')==-1:
            #q1=q1,q2=q2,width=width,minPixel=minPixel,Emin=EMin,Emax=EMax,rlu=rlu,constantBins=False,ufit=False
            ax,*_ = plotFunction(q1=self.parameters['q1'],q2=self.parameters['q2'],rlu=self.parameters['rlu'],width=self.parameters['width'],minPixel=self.parameters['width'],
            Emin=self.parameters['EMin'],Emax=self.parameters['EMax'],data=[self.pdData,self.bins],constantBins=self.parameters['constantBins'],ufit=False,label=self.name,**kwargs)
        else:
            # E1=EMin,E2=EMax,q=q1,rlu=rlu,width=width, minPixel = minPixel,ufit=False
            ax,*_ = plotFunction(q=self.parameters['q1'],rlu=self.parameters['rlu'],width=self.parameters['width'],minPixel=self.parameters['width'],
            E1=self.parameters['EMin'],E2=self.parameters['EMax'],data=[self.pdData,self.bins],ufit=False,label=self.name,**kwargs)
        return ax
    
    def save(self,location):
        self.pdData.to_csv(location)
        

    @property
    def ufit(self):
        return self._ufit
    
    @ufit.getter
    def ufit(self):
        if self._ufit is None:
            
            self._ufit = self.parameters['dataset'].generateUFitDataset(self.pdData,q1=self.parameters['q1'],
            q2=self.parameters['q2'],rlu=self.parameters['rlu'],width=self.parameters['width'],minPixel=self.parameters['width'],
            Emin=self.parameters['EMin'],Emax=self.parameters['EMax'],QDirection=self.parameters['method']=='cut1D')
        return self._ufit
        



import sys,os,json
class AppContextEmulator(object):
    def __init__(self,projectDirectory):
        self.projectDirectory = projectDirectory
        local_os = sys.platform

        if local_os.lower() == 'linux':
            local_os = 'linux'
        elif local_os.lower() == 'win32' or local_os.lower() == 'cygwin':
            local_os = 'windows'
        elif local_os.lower() == 'darwin':
            local_os = 'mac'
        else:
            local_os = 'linux'

        self.os = local_os
        self.resourses = os.path.abspath(os.path.join(os.path.dirname(__file__),'..','resources'))

        self.build_settings = {}
        self.get_build_settings()


    def get_resource(self,path):
        tryPath = os.path.join(self.resourses,'base',path)
        specificPath = os.path.join(self.resourses,self.os,path)
        
        if os.path.exists(specificPath):
            return specificPath
        else:
            return tryPath
            
    def get_build_settings(self):
        self.settingsFolder = os.path.abspath(os.path.join(os.path.dirname(__file__),'..','..','build','settings'))

        if not hasattr(self,'build_settings'):
            self.build_settings = {}

        baseFile = os.path.join(self.settingsFolder,'base.json')

        with open(baseFile) as f:
            newSettings = json.load(f)
        self.build_settings = {**self.build_settings,**newSettings}

        
        specificFile = os.path.join(self.settingsFolder,self.os+'.json')
        if os.path.exists(specificFile):
            with open(specificFile) as f:
                newSettings = json.load(f)
                self.build_settings = {**self.build_settings,**newSettings}




import sys
sys.path.append('..')

from os import path
import numpy as np

from PyQt5 import uic
import platform
from MJOLNIR._tools import WavelengthK
from MJOLNIR.TasUBlibDEG import calcCell,calculateBMatrix,calTwoTheta

try:
    from _tools import loadUI
except ModuleNotFoundError:
    from MJOLNIRGui.src.main.python._tools import loadUI

# Handles all functionality related to the CalculatorAdvancedManager. 


CalculatorAdvancedManagerBase, CalculatorAdvancedManagerForm = loadUI('Calculator_Advanced.ui')


# All of this connects the buttons and their functions to the main window.
       

def textChangedCalculate(manager):
    cell = manager.getCell()
    r = manager.getPoint()

    Cell = calcCell(cell)
    B = calculateBMatrix(Cell)

    # H,K,L,Ei,Ef
    qe = np.concatenate([r[:3],r[-2:]])

    A4 = calTwoTheta(B=B,qe=qe,ss=-1)

    qLength = np.linalg.norm(np.dot(B,np.array(r[:3])))
    d = WavelengthK(qLength)

    dE = r[-2]-r[-1]

    manager.scattering_q_label.setText('{:.4}'.format(qLength))
    manager.scattering_d_label.setText('{:.4}'.format(d))
    manager.scattering_a4_label.setText('{:.4}'.format(A4))

    manager.scattering_de_spinBox.setValue(dE)

def textChangeDeltaE(manager):
    Ef = manager.scattering_ef_spinBox.value()
    dE = manager.scattering_de_spinBox.value()

    manager.scattering_ei_spinBox.setValue(Ef+dE)




class CalculatorAdvancedManager(CalculatorAdvancedManagerBase, CalculatorAdvancedManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(CalculatorAdvancedManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow

        self.initCalculatorAdvancedManager()
        
    def initCalculatorAdvancedManager(self):    
        self.setup()
        

    def setup(self):
        # Update all boxes with check on out focus

        self.cell_a_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.cell_b_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.cell_c_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.cell_alpha_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.cell_beta_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.cell_gamma_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))

        self.scattering_ei_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.scattering_ef_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        
        self.scattering_h_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.scattering_k_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))
        self.scattering_l_spinBox.valueChanged.connect(lambda: textChangedCalculate(self))

        self.scattering_de_spinBox.valueChanged.connect(lambda: textChangeDeltaE(self))
        
    def getPoint(self):
        """Get values for scattering point"""
        Ei = getattr(self,'scattering_ei_spinBox').value()
        Ef = getattr(self,'scattering_ef_spinBox').value()

        H = getattr(self,'scattering_h_spinBox').value()
        K = getattr(self,'scattering_k_spinBox').value()
        L = getattr(self,'scattering_l_spinBox').value()

        return H,K,L,0.0,0.0,Ei,Ef



    def getCell(self):
        a = self.cell_a_spinBox.value()
        b = self.cell_b_spinBox.value()
        c = self.cell_c_spinBox.value()
        alpha = self.cell_alpha_spinBox.value()
        beta = self.cell_beta_spinBox.value()
        gamma = self.cell_gamma_spinBox.value()

        return a,b,c,alpha,beta,gamma


import sys
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments, loadUI
    import MJOLNIRGui.src.main.python._tools as _GUItools
except ImportError:
    from _tools import ProgressBarDecoratorArguments, loadUI
    import _tools as _GUItools
from os import path
from PyQt5 import QtWidgets,uic,QtGui,QtCore
import numpy as np
from MJOLNIR._tools import calculateMolarMass,symbols,_relative_atomic_masses

from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt

# Handles all functionality related to the MolecularCalculationManager. 

masses = _relative_atomic_masses.split(' ')

class ElementModel(QtCore.QAbstractTableModel):
    def __init__(self, *args, elements=None, view=None, **kwargs):
        super(ElementModel, self).__init__(*args, **kwargs)
        self.view = view
        self.reset(elements)
       
        
    def data(self, index, role):
        if role == Qt.DisplayRole:
            column = index.column()
            if column == 0:
                return self.names[index.row()]
            elif column == 1:
                return self.numbers[index.row()]
            elif column == 2:
                return self.masses[index.row()]


    def getData(self,*args,**kwargs):
         return self.data(*args,**kwargs)

    def rowCount(self, index):
        return len(self.names)

    def columnCount(self,index):
        return 3

    def reset(self,elements):
        if elements is None:
            elements = {}
        self.elements = elements
        if len(elements)==0:
            self.names = []
            self.numbers = []
            self.masses = []
        else:
            self.names = list(elements.keys())
            self.numbers = list(elements.values())

            self.masses = [masses[list(symbols).index(name)] for name in self.names]

        self.layoutChanged.emit()
        self.dataChanged.emit(self.index(0,0),self.index(len(self.names),2))

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return ["Element", "Amount", "Mass [g/mol]"][section]


    def flags(self,index):
        return QtCore.Qt.ItemIsSelectable



def checkValidSampleFormula(self,text,sampleMassL):

    try:
        mass,elements = calculateMolarMass(text,returnElements=True)
        self.setStyleSheet("color: black;")
    except AttributeError as e:
        self.setStyleSheet("color: red;")
        mass = 0
        elements = {}

    if len(text)!=0 and mass==0: # A string was provided but no mass found, typically error in input
        self.setStyleSheet("color: red;")
    elif len(text)==0:
        mass,elements = calculateMolarMass(self.placeholderText(),returnElements=True)

    sampleMassL.setText(sampleMassL.textformat.format(mass))

    self.parent().elementModel.reset(elements)


MolecularCalculationManagerBase, MolecularCalculationManagerForm = loadUI('MolecularCalculationManager.ui')


# All of this connects the buttons and their functions to the main window.

class MolecularCalculationManager(MolecularCalculationManagerBase, MolecularCalculationManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(MolecularCalculationManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/balance.png')))


        self.elementModel = ElementModel()
        self.initMolecularCalculationManager()
        
    def initMolecularCalculationManager(self):    
        
        self.setup()
        

    def setup(self):
        sampleFormLE = self.MolecularCalculationManager_sampleFormula_lineEdit
        sampleMassL = self.MolecularCalculationManager_sampleMolarMass_label

        sampleMassL.textformat = "{:.6f} g/mol"

        # Connect the check validater when text is edited
        sampleFormLE.textChanged.connect(lambda text:checkValidSampleFormula(sampleFormLE,text,sampleMassL))
        self.MolecularCalculationManager_molarMass_tableView.setModel(self.elementModel)

        # Call textChanged with empty text to update list view
        sampleFormLE.textChanged.emit('')
        





# All functionality behind the menu options of generate scripts.
import numpy as np
from os import path

from PyQt5 import QtWidgets

from MJOLNIR import _tools
from MJOLNIR.Data import Mask
import inspect
try:
    import MJOLNIRGui.src.main.python._tools as _GUItools
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments
except ImportError:
    import _tools as _GUItools
    from _tools import ProgressBarDecoratorArguments


# The general idea is to generate .py code that will load, convert and plot 
# data using the settings of the gui. Each part of the code to be written is 
# handled by different functions. At the end of the code the functions have to 
# be linked to the main window. If more functions are to be written they have 
# to be connected in MJOLNIR_GUI

def startString(ds):
    # Importing the right things
    importString = []

    importString.append('import matplotlib.pyplot as plt')
    importString.append('try:')
    importString.append('    import IPython')
    importString.append('    shell = IPython.get_ipython()')
    importString.append("    shell.enable_matplotlib(gui='qt')")
    importString.append("except:")
    importString.append("    pass\n\n")


    importString.append('from MJOLNIR import _tools')
    if not ds._maskingObject is None:
        importString.append('from MJOLNIR.Data import DataSet,Mask') 
    else:
        importString.append('from MJOLNIR.Data import DataSet') 
    
    importString.append('\nimport numpy as np')
    importString.append('from os import path')
    
    return '\n'.join(importString)

def generateFileLoader(dataFiles,name='dataFiles'):
    """Generate a string to be run in order to load the provided data file paths"""
    try:
        dataFiles = np.concatenate(dataFiles)
    except ValueError:
        pass
    
    dataFiles = np.array(dataFiles)
    
    directories = np.array([path.dirname(df) for df in dataFiles])
    
    uniqueDictionaries, indices = np.unique(directories,return_inverse = True)
        
    fileStrings = []
    
    for I, uDict in enumerate(uniqueDictionaries):
        dfs = dataFiles[indices==I]
        names = np.array([path.splitext(path.basename(df))[0] for df in dfs])
        # Find base name and remove extension
        if len(dfs) != 1:
            prefix = path.commonprefix(list(names))
            
            if 'camea' in prefix:
                # Remove all non-zero digits from prefix
                while prefix[-1]!='0' and prefix[-1]!='n':
                    prefix = prefix[:-1]
                year = int(prefix[5:9])
                numbers = np.array([n[len(prefix):] for n in names],dtype=int)
                diff = np.diff(numbers)
                separators = list(np.arange(len(diff))[np.logical_and(diff!=1,diff!=-1)]+1) # add one due to diff removing 1 lenght
                groups = []
                if len(separators) == 0:
                    groups.append('-'.join([str(numbers[0]),str(numbers[-1])]))
                else:
                    separators.insert(0,0)
                    separators.append(-1)
                    for start,stop in zip(separators[:-1],separators[1:]):
                        if stop == -1:
                            group = numbers[start:]
                        else:
                            group = numbers[start:stop]
                        if len(group)>2:
                            groups.append('-'.join([str(group[0]),str(group[-1])]))
                        elif len(group)==2:
                            groups.append(','.join(group.astype(str)))
                        else:
                            groups.append(str(group[0]))
                files = ','.join(groups)
            
                fileStrings.append('_tools.fileListGenerator("{:}",r"{:}",{:})'.format(files,uDict,year))
            else:
                fileStrings.append('["'+'",\n"'.join(dfs)+'"]')
        else:
            
            fileStrings.append('["'+dfs[0]+'"]')
            
            
    if len(fileStrings)>1:
        loadString = name+' = []\n'
        loadString+= name+'.append('+')\n'+name+'.append('.join(fileStrings)+')\n'
        loadString+= name+' = np.concatenate('+name+')'
        
    else:
        loadString = name+' = '+fileStrings[0]
    
    return loadString


def loadAndBinDataSet(ds=None,DataFiles=None,convertBeforeSubtract=None,backgroundFiles=None,binning=8,mask=None):
    """Generate string to be run in order to load a dataset"""
    

    
    loadString = []

    if binning is None:
        binString = ''
    else:
        binString = 'binning = {},'.format(binning)

    if DataFiles is None:
        loadString.append('{} = DataSet.DataSet()'.format(ds.name))
    else:
        if not backgroundFiles is None: # Dealing with subtracted data set
            loadString.append(generateFileLoader(DataFiles,name = 'foreground'))
            loadString.append('{} = DataSet.DataSet(foreground)'.format('FG'))
            loadString.append(generateFileLoader(backgroundFiles,name = 'background'))
            loadString.append('{} = DataSet.DataSet(background)'.format('BG'))

            if not convertBeforeSubtract: # Subtract first
                loadString.append('{} = FG - BG'.format(ds.name))
                loadString.append('# Run the converter. This automatically generates nxs-file(s). \n'
                                +'# Binning can be changed with binning argument.')
                loadString.append('{:}.convertDataFile({:}saveFile=False)\n\n'.format(ds.name,binString))

            else:
                loadString.append('# Run the converter. This automatically generates nxs-file(s). \n'
                                +'# Binning can be changed with binning argument.')
                
                loadString.append('{:}.convertDataFile({:}saveFile=False)\n\n'.format('FG',binString))
                loadString.append('{:}.convertDataFile({:}saveFile=False)\n\n'.format('BG',binString))
                loadString.append('{} = FG - BG'.format(ds.name))


        else:
            loadString.append(generateFileLoader(DataFiles))
            loadString.append('{} = DataSet.DataSet(dataFiles)'.format(ds.name))

            loadString.append('# Run the converter. This automatically generates nxs-file(s). \n'
                                +'# Binning can be changed with binning argument.')
            loadString.append('{:}.convertDataFile({:}saveFile=False)\n\n'.format(ds.name,binString))

    if not ds._maskingObject is None:
        loadString.append('\n')
        returnStr,totalMaskName = generateMaskStr(ds._maskingObject)
        loadString.append(returnStr)
        loadString.append('')
        loadString.append('{0}.mask = {1}({0}) # Run mask on DataSet'.format(ds.name,totalMaskName))
        loadString.append('')
        loadString.append('')


    
    
    return '\n'.join(loadString)


# Below here we have the functions for each option in the menu. 
# First the relevant parameters from the GUI are loaded, then the script is 
# generated using helper functions

@ProgressBarDecoratorArguments(runningText='Generating 3D Script',completedText='Script Saved',failedText='Cancelled')
def generate3DScript(self):
    # Check if data has been loaded. Abort if not
    self.stateMachine.run()
    if not self.stateMachine.currentState.name in ['Raw','Converted']:
        _GUItools.dialog(text='It is not possible to generate a script without any data loaded.')
        return False

    # Save the created script
    folder = self.getCurrentDirectory()
    saveFile = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File',folder,"Python (*.py)")[0]
    if len(saveFile)==0:
        return False
    if path.splitext(saveFile)[1] !='.py':
        saveFile = path.splitext(saveFile)[0]+'.py'

    # Start loading information from the GUI
    ds = self.DataSetModel.getCurrentDataSet()
    
    dataFiles = [df.original_file.fileLocation if hasattr(df,'original_file') else df.fileLocation for df in ds]
    if not ds.background is None: # Dealing with a subracted dataset
        convertBeforeSubtract = ds.convertBeforeSubtract
        backgroundFiles = ds.background
    else:
        convertBeforeSubtract = None
        backgroundFiles = None

    binning = self.ui.DataSet_binning_comboBox.currentText()
    qx = self.ui.View3D_QXBin_lineEdit.text()
    qy = self.ui.View3D_QYBin_lineEdit.text()
    E = self.ui.View3D_EBin_lineEdit.text()

    CAxisMin = self.ui.View3D_CAxisMin_lineEdit.text()
    CAxisMax = self.ui.View3D_CAxisMax_lineEdit.text()

    customViewer = self.ui.View3D_Mode_Custom_radioButton.isChecked()
    if self.ui.View3D_CurratAxe_checkBox.isChecked():
        CurratAxeBraggList = self.getBraggPoints()
    else:
        CurratAxeBraggList = None

    log = self.ui.View3D_LogScale_checkBox.isChecked()
    grid = self.ui.View3D_Grid_checkBox.isChecked()
    counts = self.ui.View3D_RawCounts_checkBox.isChecked()

    title = self.ui.View3D_SetTitle_lineEdit.text()

    RLU = self.ui.View3D_SelectUnits_RLU_radioButton.isChecked()

    # And generate the script
    generateViewer3DScript(saveFile=saveFile,dataFiles=dataFiles,ds=ds, binning=binning, qx=qx, qy=qy, E=E, 
                                RLU=RLU, CAxisMin=CAxisMin, CAxisMax=CAxisMax, log=log, grid=grid,
                                title=title,customViewer=customViewer,counts=counts,convertBeforeSubtract=convertBeforeSubtract,
                                backgroundFiles=backgroundFiles,cmap=self.colormap,CurratAxeBraggList=CurratAxeBraggList,)

    return True



def generateViewer3DScript(saveFile,ds,dataFiles,binning = None, 
                           qx=0.05, qy=0.05, E = 0.08, RLU=True, 
                           CAxisMin = 0, CAxisMax = 1e-5, log=False, grid=True, title='', selectView = 2, customViewer=False,counts=False,
                           convertBeforeSubtract=None,backgroundFiles=None,cmap='viridis',CurratAxeBraggList=None):
    # The script is generated by appending the various elements, the joining 
    # them with a \n
    saveString = []
    
    saveString.append(startString(ds=ds))
    saveString.append(loadAndBinDataSet(ds=ds,DataFiles=dataFiles,convertBeforeSubtract=convertBeforeSubtract,backgroundFiles=backgroundFiles,binning=binning))
    

    saveString.append(plotViewer3D(ds=ds, qx=qx, qy=qy, E=E, 
                                    RLU=RLU, CAxisMin=CAxisMin, CAxisMax=CAxisMax, log=log, grid=grid,
                                    title=title, selectView=selectView, customViewer=customViewer,counts=counts,cmap=cmap,CurratAxeBraggList=CurratAxeBraggList))

    saveString = '\n'.join(saveString)
    with open(saveFile,'w') as file:
        file.write(saveString)    



def plotViewer3D(ds, qx=0.05, qy=0.05, E = 0.08, RLU=True, 
                CAxisMin = 0, CAxisMax = 1e-5, log=False, grid=True, title='', selectView = 2, customViewer=False,counts=False,cmap='viridis',CurratAxeBraggList=None):
    # Generate the actual text for the View3D script

    plotString = []

    # Check which arguments have been used in the GUI. If they are unused or 
    # left to their default values we do not want the argumetns to clutter the generated script.
    if RLU == False and customViewer == False:
        rluArgument = ',rlu = False'
    else:
        rluArgument = ''
    
    if log == True:
        logArgument = ', log=True'
    else:
        logArgument = ''

    if grid == True:
        gridArgument = ', grid=True'
    else:
        gridArgument = ', grid=False'

    if customViewer == True:
        sliceArgument = ', customSlicer = True'
    else:
        sliceArgument = ''

    if counts == True and customViewer == False:
        countsArgument = ', counts = True'
    else:
        countsArgument = ''

    if cmap != 'viridis':
        cmapArgument = ', cmap = "'+cmap+'"'
    else:
        cmapArgument = ''

    if not CurratAxeBraggList is None:
        plotString.append('\n# Make list of Bragg peaks used for Currat-Axe spurion calculation\nCurratAxeBraggList = {}\n\n'.format(CurratAxeBraggList))
        CABraggArgument = ', CurratAxeBraggList=CurratAxeBraggList'
    else:
        CABraggArgument = ''

    args = rluArgument+logArgument+gridArgument+sliceArgument+countsArgument+cmapArgument+CABraggArgument

    plotString.append('# Plotting data quickly in equi-sized voxels can be done by')
    plotString.append('Viewer = {}.View3D({}{})'.format(ds.name,','.join([str(x) for x in [qx,qy,E]]),args))

    plotString.append("# Above, all data is binned in voxels of size " +str(qx) + "/AA, " + str(qy) +"/AA, and " +str(E) +" meV.\n"\
                        +"# Automatically, data is plotted in reciprocal lattice as provided by the\n"\
                        +"# UB matrix saved in the data file. When using the custom viewer, only RLU is supported.\n"\
                        +"#Alternatively, one can plot in 'raw' "\
                        +"coordinates (orthogonal and in units of 1/AA)using the standard viewer \n#by issuing rlu=False above.\n")
    
    plotString.append('Viewer.caxis=({},{})'.format(CAxisMin,CAxisMax))

    plotString.append("# Without any intervention data is usually plotted on a useless colour scale.\n"\
                        +"# This is countered by specifying min and max for colour by the call above.\n"\
                        +"# Alternatively, one can provide this as input to View3D\n")

    if title !='':
        plotString.append('# Set title of plot')
        plotString.append('Viewer.set_title("{}")\n'.format(title))

    if selectView !=2 and customViewer==False:
        plotString.append('# Set View direction')
        plotString.append('Viewer.setProjection({})\n'.format(selectView))

    plotString.append('plt.show()\n')
        
    return '\n'.join(plotString)


@ProgressBarDecoratorArguments(runningText='Generating QE Script',completedText='Script Saved',failedText='Cancelled')
def generateQEScript(self):
    self.stateMachine.run()
    if not self.stateMachine.currentState.name in ['Raw','Converted']:
        _GUItools.dialog(text='It is not possible to generate a script without any data loaded.')
        return False

    folder = self.getCurrentDirectory()
    saveFile = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File',folder,"Python (*.py)")[0]
    if len(saveFile)==0:
        return False
    if path.splitext(saveFile)[1] !='.py':
        saveFile = path.splitext(saveFile)[0]+'.py'

    ds = self.DataSetModel.getCurrentDataSet()
    
    dataFiles = [df.original_file.fileLocation if hasattr(df,'original_file') else df.fileLocation for df in ds]
    if not ds.background is None: # Dealing with a subracted dataset
        convertBeforeSubtract = ds.convertBeforeSubtract
        backgroundFiles = ds.background
    else:
        convertBeforeSubtract = None
        backgroundFiles = None

    binning = self.ui.DataSet_binning_comboBox.currentText()
    
    HStart = self.ui.QELine_HStart_lineEdit.text()
    KStart = self.ui.QELine_KStart_lineEdit.text()
    LStart = self.ui.QELine_LStart_lineEdit.text()
    HEnd = self.ui.QELine_HEnd_lineEdit.text()
    KEnd = self.ui.QELine_KEnd_lineEdit.text()
    LEnd = self.ui.QELine_LEnd_lineEdit.text()
    width = self.ui.QELine_Width_lineEdit.text()
    minPixel = self.ui.QELine_MinPixel_lineEdit.text()
    EMin = self.ui.QELine_EMin_lineEdit.text()
    EMax = self.ui.QELine_EMax_lineEdit.text()
    NPoints = self.ui.QELine_NPoints_lineEdit.text()

    constantBins = self.ui.QELine_ConstantBins_checkBox.isChecked()

    CAxisMin = self.ui.QELine_CAxisMin_lineEdit.text()
    CAxisMax = self.ui.QELine_CAxisMax_lineEdit.text()

    log = self.ui.QELine_LogScale_checkBox.isChecked()
    grid = self.ui.QELine_Grid_checkBox.isChecked()

    title = self.ui.QELine_SetTitle_lineEdit.text()

    RLU = self.ui.QELine_SelectUnits_RLU_radioButton.isChecked()
    
    
    generatePlotQEScript(saveFile=saveFile,ds=ds,dataFiles=dataFiles,binning = binning, 
                                HStart=HStart, KStart=KStart, LStart = LStart, HEnd=HEnd, KEnd=KEnd, LEnd=LEnd, 
                                width=width, minPixel=minPixel, EMin = EMin, EMax=EMax, NPoints=NPoints, RLU=RLU, 
                                CAxisMin = CAxisMin, CAxisMax = CAxisMax, log=log, grid=grid, title=title, constantBins=constantBins,
                                convertBeforeSubtract=convertBeforeSubtract,backgroundFiles=backgroundFiles,cmap=self.colormap)

    return True    

    
def generatePlotQEScript(saveFile,ds,dataFiles,binning = None, 
                             HStart=-1, KStart=0.0, LStart = -1, HEnd=-1, KEnd=0, LEnd=1,
                             width=0.1, minPixel=0.01, EMin = 0.0, EMax=10, NPoints=101, RLU=True, 
                             CAxisMin = 0, CAxisMax = 1e-5, log=False, grid=True, title='', constantBins=False,
                             convertBeforeSubtract=None,backgroundFiles=None,cmap='viridis'):
    saveString = []
    
    saveString.append(startString(ds=ds))
    saveString.append(loadAndBinDataSet(ds=ds,DataFiles=dataFiles,convertBeforeSubtract=convertBeforeSubtract,backgroundFiles=backgroundFiles,binning=binning))

    saveString.append(plotQEText(ds=ds, HStart=HStart, KStart=KStart, LStart = LStart, HEnd=HEnd, KEnd=KEnd, LEnd=LEnd,
               width=width, minPixel=minPixel, EMin = EMin,EMax=EMax,NPoints=NPoints, RLU=RLU, 
                CAxisMin = CAxisMin, CAxisMax = CAxisMax, log=log, grid=grid, title=title, constantBins=constantBins,cmap=cmap))

    saveString = '\n'.join(saveString)    
        
    with open(saveFile,'w') as file:
        file.write(saveString)
        

def plotQEText(ds, HStart=-1, KStart=0.0, LStart = -1, HEnd=-1, KEnd=0, LEnd=1,
               width=0.1, minPixel=0.01, EMin = 0.0,EMax=10,NPoints=101, RLU=True, 
                CAxisMin = 0, CAxisMax = 1e-5, log=False, grid=True, title='', constantBins=False,cmap='viridis'):

    plotString = []

    if RLU == False:
        rluArgument = ',rlu = False'
    else:
        rluArgument = ''
    
    if log == True:
        logArgument = ', log=True'
    else:
        logArgument = ''

    if cmap != 'viridis':
        cmapArgument = ', cmap = "'+cmap+'"'
    else:
        cmapArgument = ''

    args = rluArgument+logArgument+cmapArgument

    plotString.append('# Plotting a cut through data in Q and E is done using this code')
    
    
    plotString.append('# First define the positions to be cut through.')
    plotString.append('# The GUI is limited to two Q points, but more can be added below')
    
    plotString.append('Q1 = np.array([' + HStart + ',' + KStart + ',' + LStart + '])')
    plotString.append('Q2 = np.array([' + HEnd + ',' + KEnd + ',' + LEnd + '])')
    
    plotString.append('# Define orthogonal width and minimum pixel size along Q-cut')
    plotString.append('width = ' + width + ' # 1/AA')
    plotString.append('minPixel = ' + minPixel + ' # 1/AA')
    
    plotString.append('# Define energy bins')
    plotString.append('EnergyBins=np.linspace(' + EMin + ',' + EMax + ',' + NPoints +')')  
        
    
    plotString.append("ax,Data,Bins = \\" )
    plotString.append('        ' + ds.name + '.plotCutQE(q1=Q1, q2=Q2, width=width, minPixel=minPixel, \\')
    plotString.append('        ' +'EnergyBins=EnergyBins' +args +')')

    plotString.append("# Without any intervention data is usually plotted on a useless colour scale.\n"\
                        +"# This is countered by specifying min and max for colour by the call below.\n"\
                        +"# Alternatively, one can provide this as input to plotCutQELine\n")

    plotString.append('ax.set_clim(' + CAxisMin + ',' + CAxisMax +')')

    if grid == True:
        plotString.append('ax.grid(True,zorder=0,c=\'k\') ')


    if title !='':
        plotString.append('# Set title of plot')
        plotString.append('ax.set_title("{}")\n'.format(title))

    plotString.append('plt.show()\n')
        
    return '\n'.join(plotString)

        

@ProgressBarDecoratorArguments(runningText='Generating QPlane Script',completedText='Script Saved',failedText='Cancelled')
def generateQPlaneScript(self):
    self.stateMachine.run()
    if not self.stateMachine.currentState.name in ['Raw','Converted']:
        _GUItools.dialog(text='It is not possible to generate a script without any data loaded.')
        return False

    folder = self.getCurrentDirectory()
    saveFile = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File',folder,"Python (*.py)")[0]
    if len(saveFile)==0:
        return False
    if path.splitext(saveFile)[1] !='.py':
        saveFile = path.splitext(saveFile)[0]+'.py'

    ds = self.DataSetModel.getCurrentDataSet()
    
    dataFiles = [df.original_file.fileLocation if hasattr(df,'original_file') else df.fileLocation for df in ds]
    if not ds.background is None: # Dealing with a subracted dataset
        convertBeforeSubtract = ds.convertBeforeSubtract
        backgroundFiles = ds.background
    else:
        convertBeforeSubtract = None
        backgroundFiles = None

    binning = self.ui.DataSet_binning_comboBox.currentText()
    
    
    
    EMin = self.ui.QPlane_EMin_lineEdit.text()
    EMax = self.ui.QPlane_EMax_lineEdit.text()

    xBinTolerance = self.ui.QPlane_xBinTolerance_lineEdit.text()
    yBinTolerance = self.ui.QPlane_yBinTolerance_lineEdit.text()

    CAxisMin = self.ui.QPlane_CAxisMin_lineEdit.text()
    CAxisMax = self.ui.QPlane_CAxisMax_lineEdit.text()

    log = self.ui.QPlane_LogScale_checkBox.isChecked()
    grid = self.ui.QPlane_Grid_checkBox.isChecked()

    title = self.ui.QPlane_SetTitle_lineEdit.text()

    RLU = self.ui.QPlane_SelectUnits_RLU_radioButton.isChecked()
    
    
    generatePlotQPlaneScript(saveFile,ds,dataFiles,binning = binning, 
                             xBinTolerance=xBinTolerance, yBinTolerance=yBinTolerance,
                             EMin = EMin,EMax=EMax, RLU=RLU, 
                             CAxisMin = CAxisMin, CAxisMax = CAxisMax, log=log, grid=grid, title=title,
                             convertBeforeSubtract=convertBeforeSubtract,backgroundFiles=backgroundFiles,cmap=self.colormap)

    return True    


def generatePlotQPlaneScript(saveFile,ds,dataFiles,binning = None, 
                             xBinTolerance=0.03, yBinTolerance=0.03,
                             EMin = 0.0,EMax=10, RLU=True, 
                             CAxisMin = 0, CAxisMax = 1e-5, log=False, grid=True, title='',
                             convertBeforeSubtract=None,backgroundFiles=None,cmap='viridis'):
    saveString = []
    
    saveString.append(startString(ds=ds))
    saveString.append(loadAndBinDataSet(ds=ds,DataFiles=dataFiles,convertBeforeSubtract=convertBeforeSubtract,backgroundFiles=backgroundFiles,binning=binning))
    
    saveString.append(plotQPlaneText(ds=ds, xBinTolerance=xBinTolerance, yBinTolerance=yBinTolerance,
                EMin = EMin,EMax=EMax, RLU=RLU, 
                CAxisMin = CAxisMin, CAxisMax = CAxisMax, log=log, grid=grid, title=title,cmap=cmap))

    saveString = '\n'.join(saveString)    
        
    with open(saveFile,'w') as file:
        file.write(saveString)


def plotQPlaneText(ds, xBinTolerance=0.03, yBinTolerance=0.03,
                EMin = 0.0,EMax=10, RLU=True, 
                CAxisMin = 0, CAxisMax = 1e-5, log=False, grid=True, title='',cmap='viridis'):

    plotString = []

    if RLU == False:
        rluArgument = ',rlu = False'
    else:
        rluArgument = ''
    
    if log == True:
        logArgument = ', log=True'
    else:
        logArgument = ''

    if cmap != 'viridis':
        cmapArgument = ', cmap = "'+cmap+'"'
    else:
        cmapArgument = ''
    args = rluArgument+logArgument+cmapArgument

    plotString.append('# Plotting a Q plane done using this code')

    plotString.append('EMin = ' + EMin)
    plotString.append('EMax = ' + EMax)
    plotString.append('xBinTolerance = ' + xBinTolerance)
    plotString.append('yBinTolerance = ' + yBinTolerance)
    plotString.append('Data,Bins,ax = '+ ds.name + '.plotQPlane(EMin=EMin, EMax=EMax,xBinTolerance=xBinTolerance,yBinTolerance=yBinTolerance' + args + ')')

    plotString.append('fig = ax.get_figure() # Extract figure from returned axis')
    plotString.append('fig.colorbar(ax.pmeshs[0]) # Create colorbar from plot\n')

    plotString.append("# Without any intervention data is usually plotted on a useless colour scale.\n"\
                        +"# This is countered by specifying min and max for colour by the call below.\n"\
                        +"# Alternatively, one can provide this as input to plotCutQPlane\n")

    plotString.append('ax.set_clim(' + CAxisMin + ',' + CAxisMax +')')

    if grid == True:
        plotString.append('ax.grid(True,zorder=0) ')


    if title !='':
        plotString.append('# Set title of plot')
        plotString.append('ax.set_title("{}")\n'.format(title))

    plotString.append('plt.show()\n')
        
    return '\n'.join(plotString)



@ProgressBarDecoratorArguments(runningText='Generating Cut1D Script',completedText='Script Saved',failedText='Cancelled')
def generateCut1DScript(self):
    self.stateMachine.run()
    if not self.stateMachine.currentState.name in ['Raw','Converted']:
        _GUItools.dialog(text='It is not possible to generate a script without any data loaded.')
        return False

    if self.Cut1DModel.rowCount() == 0:
        _GUItools.dialog(text='It is not possible to generate a script without any cuts performed. Please make a cut using the "Generate 1D cut" or Plot 1D cut" buttons.')
        return False

    folder = self.getCurrentDirectory()
    saveFile = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File',folder,"Python (*.py)")[0]
    if len(saveFile)==0:
        return False
    if path.splitext(saveFile)[1] !='.py':
        saveFile = path.splitext(saveFile)[0]+'.py'

    ds = self.DataSetModel.getCurrentDataSet()
    
    dataFiles = [df.original_file.fileLocation if hasattr(df,'original_file') else df.fileLocation for df in ds]
    if not ds.background is None: # Dealing with a subracted dataset
        convertBeforeSubtract = ds.convertBeforeSubtract
        backgroundFiles = ds.background
    else:
        convertBeforeSubtract = None
        backgroundFiles = None

    binning = self.ui.DataSet_binning_comboBox.currentText()
    
    saveString = []
    saveString.append(startString(ds=ds))
    saveString.append('\n\n')
    saveString.append(loadAndBinDataSet(ds=ds,DataFiles=dataFiles,convertBeforeSubtract=convertBeforeSubtract,backgroundFiles=backgroundFiles,binning=binning))
    self.appendPLTSHOW = False # Flag to keep track of the need for a plt.show() command
    if self.Cut1DModel.rowCount() == 1:
        cut = self.Cut1DModel.dataCuts1D[0]
        saveString.append(plotCut1DText(ds,cut,single=True,self=self))
    else:
        
        for cut in self.Cut1DModel.dataCuts1D:
            saveString.append(plotCut1DText(ds,cut,single=False,self=self))
            saveString.append('\n')

    if self.appendPLTSHOW:
        saveString.append('\nplt.show()')
    saveString.append('\n\n#If a ufit object is needed, add "ufit=True" to the above method calls and change "_data" and "_bins" to "_ufit".')
    with open(saveFile,'w') as file:
        file.write(''.join(saveString))
    return True    
        
        
def plotCut1DText(ds, cut,single=True,self=None):
    title = cut.name.replace(' ','_')
    if cut.parameters['method'].find('cut1DE')>-1: # If the method contains "cut1DE" it is for constant q
        q = cut.parameters['q1']
        EMin,EMax = cut.parameters['EMin'],cut.parameters['EMax']
        rlu,width = cut.parameters['rlu'],cut.parameters['width']
        minPixel = cut.parameters['minPixel']
        constantBins = cut.parameters['constantBins']
        ufit = not cut._ufit is None
        method = cut.parameters['method']

        if cut.parameters['method'].find('plot')>-1:
            returnPars = title+'_ax,'
        else:
            returnPars = ''
        if ufit:
            returnPars+=title+'_cut'
        else:
            returnPars+=title+'_data,'+title+'_bins'

        plotString = []
        if single:
            plotString.append('# Plotting a 1D cut for constant q through data is done using this code')
            plotString.append('# First define the positions to be cut through.')
            
            plotString.append('q = np.array(['+','.join([str(x) for x in q])+ '])')
            
            plotString.append('# Define orthogonal width and minimum pixel size along Q-cut')
            plotString.append('width = ' + str(width) + ' # 1/AA')
            plotString.append('minPixel = ' + str(minPixel) + ' # 1/AA')
            plotString.append('rlu = ' + str(rlu))
            plotString.append('constantBins = ' + str(constantBins))

            plotString.append('# Define energies')
            plotString.append('E1='+str(EMin))
            plotString.append('E2='+str(EMax))
            
            plotString.append(returnPars+' = ' +ds.name +'.'+method+'(E1=E1,E2=E2,q=q,width=width,minPixel=minPixel,rlu=rlu,constantBins=constantBins,ufit='+str(ufit)+')')
        else:
            plotString.append(returnPars+' = ' +ds.name +'.'+method+'(E1={},E2={},q={},width={},minPixel={},rlu={},constantBins={},ufit={})'.format(
            EMin,EMax,'['+','.join([str(x) for x in q])+ ']',width,minPixel,rlu,constantBins,ufit))
        
    else:
        q1,q2 = cut.parameters['q1'], cut.parameters['q2']
        EMin,EMax = cut.parameters['EMin'],cut.parameters['EMax']
        rlu,width = cut.parameters['rlu'],cut.parameters['width']
        minPixel = cut.parameters['minPixel']
        constantBins = cut.parameters['constantBins']
        ufit = not cut._ufit is None
        method = cut.parameters['method']

        plotString = []

        if method.find('plot')>-1:
            returnPars = title+'_ax,'
        else:
            returnPars = ''
        if ufit:
            returnPars+=title+'_cut'
        else:
            returnPars+=title+'_data,'+title+'_bins'

        if single:
            plotString.append('# Plotting a 1D cut through data is done using this code')
            
            
            plotString.append('# First define the positions to be cut through.')
            
            plotString.append('Q1 = np.array(['+','.join([str(x) for x in q1])+ '])')
            plotString.append('Q2 = np.array(['+','.join([str(x) for x in q2])+ '])')
            
            plotString.append('# Define orthogonal width and minimum pixel size along Q-cut')
            plotString.append('width = ' + str(width) + ' # 1/AA')
            plotString.append('minPixel = ' + str(minPixel) + ' # 1/AA')
            plotString.append('rlu = ' + str(rlu))
            plotString.append('constantBins = ' + str(constantBins))

            plotString.append('# Define energies')
            plotString.append('EMin='+str(EMin))
            plotString.append('EMax='+str(EMax))

            plotString.append(returnPars+' = ' +ds.name +'.'+method+'(q1=Q1,q2=Q2,width=width,minPixel=minPixel,Emin=EMin,Emax=EMax,rlu=True,constantBins=constantBins,ufit='+str(ufit)+')')
        else:
            plotString.append(returnPars+' = ' +ds.name +'.'+method+'(q1={},q2={},width={},minPixel={},Emin={},Emax={},rlu={},constantBins={},ufit={})'.format(
                '['+','.join([str(x) for x in q1])+ ']','['+','.join([str(x) for x in q2])+ ']',width,minPixel,EMin,EMax,rlu,constantBins,ufit
            ))

    if method.find('plot')>-1:
        plotString.append('# Set title of plot')
        plotString.append(title+'_ax.set_title("{}")\n'.format(cut.name))
        self.appendPLTSHOW = True

        
    return '\n'.join(plotString)

# Teach the main window how to call the functions in here.
def initGenerateScript(guiWindow):
    guiWindow.generateQEScript = lambda: generateQEScript(guiWindow)
    guiWindow.generate3DScript = lambda: generate3DScript(guiWindow)
    guiWindow.generateQPlaneScript = lambda: generateQPlaneScript(guiWindow)
    guiWindow.generateCut1DScript = lambda: generateCut1DScript(guiWindow)

def setupGenerateScript(guiWindow):
    pass

def generateMaskStr(maskingObject):
    
    def generateMask(mask):
        makeStr = []
        signature = inspect.signature(type(mask).__init__)
        
        paramDict = {}
        maskName = mask.name
        for param in signature.parameters:
            if param == 'self': continue
            paramDict[param] = getattr(mask,param)
            if not isinstance(paramDict[param],(np.ndarray)):
                if isinstance(paramDict[param],(str)):
                    makeStr.append(maskName+'_'+param+' = "'+str(paramDict[param])+'"')
                else:
                    makeStr.append(maskName+'_'+param+' = '+str(paramDict[param]))
            else:
                makeStr.append(maskName+'_'+param+' = np.array('+np.array2string(paramDict[param],separator=',')+')')
            
        makeStr.append('')
        makeStr.append(paramDict['name']+' = Mask.'+type(mask).__name__+'(')
        for idx,key in enumerate(paramDict):
            makeStr.append('\t'+key+' = '+maskName+'_'+key+','*(idx!=len(paramDict)-1)+')'*(idx==len(paramDict)-1))
    
        return '\n'.join(makeStr)
    
    eq,masks = Mask.extract(maskingObject)
    masks.sort(key=lambda mask:mask.name)
    defineMaskStr = ['# Generate individual masks']+[generateMask(mask) for mask in masks]
    totalMaskName = 'totalMask'
    while totalMaskName in [m.name for m in masks]:
        try:
            idx = int(totalMaskName.split('_')[-1])
        except ValueError:
            totalMaskName = totalMaskName+'_1'
        else:
            totalMaskName = totalMaskName[:-(len(str(idx))+1)] + '_' + str(idx+1)
    
    maskStr = '\n\n'.join(['\n\n'.join(defineMaskStr),totalMaskName+' = '+eq])

    return maskStr,totalMaskName


import sys
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import ProgressBarDecoratorArguments,loadUI
    from MJOLNIRGui.src.main.python.Views import BraggListManager
    import MJOLNIRGui.src.main.python._tools as _GUItools
    from MJOLNIRGui.src.main.python.DataModels import MatplotlibFigureList,MatplotlibFigureListDelegate
    from MJOLNIRGui.src.main.python.MJOLNIR_Data import Gui1DCutObject
except ImportError:
    from _tools import ProgressBarDecoratorArguments,loadUI
    from Views import BraggListManager
    import _tools as _GUItools
    from DataModels import MatplotlibFigureList,MatplotlibFigureListDelegate
    from MJOLNIR_Data import Gui1DCutObject
from os import path
from PyQt5 import QtWidgets,uic
import numpy as np
from MJOLNIR.Data import Viewer3D
from MJOLNIR.Data.DraggableShapes import extractCut1DPropertiesRectangle, extractCut1DPropertiesCircle
import matplotlib.pyplot as plt

# Handles all functionality related to the View3D box. Each button has its own 
# definition, which should be pretty selfexplanatory.

def View3D_setCAxis_button_function(self):
    currentFigure = self.figureListView3D.getCurrentFigure()
    if currentFigure is None:
        self.View3D_plot_button_function()
        currentFigure = self.figureListView3D.getCurrentFigure()
        
    CAxisMin=float(self.ui.View3D_CAxisMin_lineEdit.text())
    CAxisMax=float(self.ui.View3D_CAxisMax_lineEdit.text())
            
    currentFigure.set_clim(CAxisMin,CAxisMax)
    if hasattr(currentFigure,'settings'):
        currentFigure.settings['View3D_CAxisMax_lineEdit'] = str(CAxisMax)
        currentFigure.settings['View3D_CAxisMin_lineEdit'] = str(CAxisMin)
    


def View3D_SetTitle_button_function(self):        
    currentFigure = self.figureListView3D.getCurrentFigure()
    if not currentFigure is None:
        TitleText=self.ui.View3D_SetTitle_lineEdit.text()
        if TitleText == '':
            TitleText = self.ui.View3D_SetTitle_lineEdit.placeholderText()
        if isinstance(currentFigure,Viewer3D.Viewer3D):
            currentFigure.ax.set_title(TitleText)
            
            # Get the value of the slider right now, then change it around a bit and put it back to where it was, to render properly
            currentSliderValue=currentFigure.Energy_slider.val
            currentFigure.Energy_slider.set_val(0)
            currentFigure.Energy_slider.set_val(1)
            currentFigure.Energy_slider.set_val(currentSliderValue)
        else:
            currentFigure.parent().window().setWindowTitle(TitleText)

        currentFigure._title = TitleText
        
        idx = self.figureListView3D.getCurrentFigureIndex()
        dummyIdx = self.figureListView3D.index(idx)
        self.figureListView3D.beginInsertRows(dummyIdx,idx,idx+1)
        self.figureListView3D.endInsertRows()
        self.figureListView3D.view.currentTextChanged.emit(TitleText)
        if hasattr(currentFigure,'settings'):
            currentFigure.settings['View3D_SetTitle_lineEdit'] = TitleText
        
                
@ProgressBarDecoratorArguments(runningText='Generating View3D',completedText='View3D Generated')                    
def View3D_plot_button_function(self):
    if not self.stateMachine.requireStateByName('Converted'):
        return False

    # Check if we already have data, otherwise convert current data.
    ds = self.DataSetModel.getCurrentDataSet()
    if len(ds.convertedFiles)==0:
        self.DataSet_convertData_button_function()

    # Extract all the information from the GUI
    QXBin=float(self.ui.View3D_QXBin_lineEdit.text())
    QYBin=float(self.ui.View3D_QYBin_lineEdit.text())
    EBin =float(self.ui.View3D_EBin_lineEdit.text())
    
    if self.ui.View3D_SelectUnits_RLU_radioButton.isChecked():
        rlu=True
        
    if self.ui.View3D_SelectUnits_AA_radioButton.isChecked():
        rlu=False

    if self.ui.View3D_Grid_checkBox.isChecked():
        grid=9
    else:
        grid=False
    
    log = self.ui.View3D_LogScale_checkBox.isChecked()
    customSlicer = not self.ui.View3D_Mode_Viewer3D_radioButton.isChecked()
    counts = self.ui.View3D_RawCounts_checkBox.isChecked()
    plotCurratAxe = self.ui.View3D_CurratAxe_checkBox.isChecked()
    braggList = self.getBraggPoints()

    cut1DFunctionRectangleLocal = lambda viewer,dr:cut1DFunctionRectangle(self,viewer,dr)    
    cut1DFunctionCircleLocal = lambda viewer,dr:cut1DFunctionCircle(self,viewer,dr)    
    figure = ds.View3D(QXBin,QYBin,EBin,grid=grid,rlu=rlu,log=log,customSlicer=customSlicer,counts=counts,
    outputFunction=self.writeToStatus,cmap=self.colormap,CurratAxeBraggList=braggList,plotCurratAxe=plotCurratAxe,cut1DFunctionRectangle=cut1DFunctionRectangleLocal,
    cut1DFunctionCircle=cut1DFunctionCircleLocal)
    self.figureListView3D.append(figure)
    currentFigure = self.figureListView3D.getCurrentFigure()

    TitleText=self.ui.View3D_SetTitle_lineEdit.text()
    if TitleText == '':
        TitleText = self.ui.View3D_SetTitle_lineEdit.placeholderText()

    CAxisMin=float(self.ui.View3D_CAxisMin_lineEdit.text())
    CAxisMax=float(self.ui.View3D_CAxisMax_lineEdit.text())

    figure.settings = {'View3D_SelectUnits_RLU_radioButton':rlu,
                       'View3D_SelectUnits_AA_radioButton': not rlu,
                       'View3D_Grid_checkBox':grid!=False,
                       'View3D_LogScale_checkBox':log,
                       'View3D_Mode_Viewer3D_radioButton': not customSlicer,
                       'View3D_Mode_Custom_radioButton': customSlicer,
                       'View3D_RawCounts_checkBox':counts,
                       'View3D_CurratAxe_checkBox':plotCurratAxe,
                       'View3D_QXBin_lineEdit':QXBin,
                       'View3D_QYBin_lineEdit':QYBin,
                       'View3D_EBin_lineEdit':EBin,
                       'View3D_SetTitle_lineEdit':TitleText,
                       'View3D_CAxisMax_lineEdit':CAxisMax,
                       'View3D_CAxisMin_lineEdit':CAxisMin}

    if customSlicer:
        #self.windows.append(currentFigure.parent())
        def setClosed(self,fig):
            self.figureListView3D.close(fig)
            fig.closed=True

        closeFunction = lambda event: setClosed(self,currentFigure)
        currentFigure.parent().closeEvent = closeFunction
    else:
        #self.windows.append(currentFigure.ax.get_figure())
        def setClosed(fig):
            fig.closed=True

        closeFunction = lambda event: setClosed(currentFigure)
        
        currentFigure.ax.get_figure().canvas.mpl_connect('close_event', closeFunction)
    
        self.View3D_SetTitle_button_function()
        currentFigure.setPlane(1)
        currentFigure.setPlane(0)

    
    self.View3D_setCAxis_button_function()      
    
    
    return True
    #except:
    #    _GUItools.dialog(text='View3D plot could not be made. Check the input parameters and try again!')
    #    return False        
    
def View3D_toggle_mode_function(self):
    if self.ui.View3D_Mode_Viewer3D_radioButton.isChecked(): # changed to Viewer3D
        # Change titles
        self.ui.View3D_Grid_checkBox.setEnabled(True)

        self.ui.View3D_SelectUnits_RLU_radioButton.setEnabled(True)
        self.ui.View3D_SelectUnits_AA_radioButton.setEnabled(True)

    else: # Changing to AA
        self.ui.View3D_Grid_checkBox.setEnabled(True)

        self.ui.View3D_SelectUnits_RLU_radioButton.setEnabled(False)
        self.ui.View3D_SelectUnits_AA_radioButton.setEnabled(False)
        
def View3D_toggle_plotCurratAxe_function(self):
    currentFigure = self.figureListView3D.getCurrentFigure()
    if not currentFigure is None:
        value = self.ui.View3D_CurratAxe_checkBox.isChecked()
        currentFigure.plotCurratAxe = value
        if hasattr(currentFigure,'settings'):
            currentFigure.settings['View3D_CurratAxe_checkBox'] = value

def View3D_changed_CurratAxeList_function(self):
    currentFigure = self.figureListView3D.getCurrentFigure()
    if not currentFigure is None:
        braggPoints =  self.getBraggPoints()
        self.ui.View3D_CurratAxe_checkBox.setChecked(True)
        currentFigure.CurratAxeBraggList = braggPoints

        

def View3D_DataSet_selectionChanged_function(self):
    ds = self.DataSetModel.getCurrentDataSet()
    if not ds is None:
        title = ds.name
    else:
        title = ''
    self.ui.View3D_SetTitle_lineEdit.setPlaceholderText(title)
    
def View3D_Grid_checkBox_toggled_function(self):
    currentFigure = self.figureListView3D.getCurrentFigure()
    if not currentFigure is None:
        if self.ui.View3D_Grid_checkBox.isChecked():
            currentFigure.grid = True
            currentFigure.gridZOrder=9
        else:
            currentFigure.grid = False
            currentFigure.gridZOrder=9
        currentFigure.settings['View3D_Grid_checkBox'] = currentFigure.grid!=False
        for ax in currentFigure._axes:
            ax.grid(currentFigure.grid,zorder=currentFigure.gridZOrder)

def indexChanged(self,index):
    figure = self.figureListView3D.figures[index]
    if hasattr(figure,'settings'):
        for setting,value in figure.settings.items():
            if 'radio' in setting or 'checkBox' in setting:
                getattr(getattr(self.ui,setting),'setChecked')(value)
            else:
                getattr(getattr(self.ui,setting),'setText')(str(value))
    if not figure is None:
        self.ui.View3D_figureToCSV_pushButton.setEnabled(True)
    else:
        self.ui.View3D_figureToCSV_pushButton.setEnabled(False)

@ProgressBarDecoratorArguments(runningText='Save View3D to csv',completedText='View3D saved')
def saveToCSV(self):
    figure = self.figureListView3D.getCurrentFigure()
    ax = figure
    try:
        title = ax.ax.set_title.get_text()
    except AttributeError:
        title = 'View3D'
    location,_ = QtWidgets.QFileDialog.getSaveFileName(self,'Save View3D',str(title)+'.csv','CSV (*.csv)')
    if location is None or location == '':
        return False
    if not location.split('.')[-1] == 'csv':
        location = location+'.csv'
    
    ax.to_csv(location)
    return True

## Function to connect draggable rectangles from interactive 1D cuts to the gui
def cut1DFunctionRectangle(self,viewer,dr):
    # If there is no sample, i.e. 1/AA plot
    if hasattr(viewer.ax,'sample'):
        sample = viewer.ax.sample
    else:
        sample = None

    # Extract the parameters
    rounding = 4 # Round to 4 digits
    parameters = extractCut1DPropertiesRectangle(dr.rect,sample,rounding = rounding)
    step = viewer.dQE[viewer.axis]
    pos = np.zeros(3,dtype=int)
    pos[viewer.axis]=viewer.Energy_slider.val
    
    # Find the correct energy bin for the current plot
    EMin = np.round(viewer.bins[viewer.axis][pos[0],pos[1],pos[2]],rounding)
    EMax = np.round(EMin+step,rounding)
    
    parameters['Emin']=np.round(EMin,rounding)
    parameters['Emax']=np.round(EMax,rounding)

    # Order of parameters needed is: ds,q1,q2,width,minPixel,EMax,EMin,cutQ,rlu
    self.interactiveCut = [viewer.ds,parameters['q1'],parameters['q2'],
                           parameters['width'],parameters['minPixel'],EMax,EMin,True,parameters['rlu']]
    
    # Perform the cut and plot it
    self.Cut1D_plot_button_function()
    # Reset the interactiveCut flag
    self.interactiveCut = None
    
## Function to connect draggable rectangles from interactive 1D cuts to the gui
def cut1DFunctionCircle(self,viewer,dr):
    # If there is no sample, i.e. 1/AA plot
    if hasattr(viewer.ax,'sample'):
        sample = viewer.ax.sample
    else:
        sample = None

    # Extract the parameters
    rounding = 4 # Round to 4 digits
    parameters = extractCut1DPropertiesCircle(dr.circ,sample,rounding = rounding)
    parameters['minPixel'] = np.round(viewer.dQE[viewer.axis],rounding)
    
    
    parameters['Emin']=np.round(viewer.ds.energy.min(),rounding)
    parameters['Emax']=np.round(viewer.ds.energy.max(),rounding)

    # Order of parameters needed is: ds,q1,q2,width,minPixel,EMax,EMin,cutQ,rlu
    self.interactiveCut = [viewer.ds,parameters['q'],None,
                           parameters['width'],parameters['minPixel'],parameters['Emax'],parameters['Emin'],False,parameters['rlu']]
    
    # Perform the cut and plot it
    self.Cut1D_plot_button_function()
    # Reset the interactiveCut flag
    self.interactiveCut = None

View3DManagerBase, View3DManagerForm = loadUI('View3D.ui')

class View3DManager(View3DManagerBase, View3DManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(View3DManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.initView3DManager()
        
    def initView3DManager(self):    
        self.guiWindow.View3D_setCAxis_button_function = lambda:View3D_setCAxis_button_function(self.guiWindow)
        self.guiWindow.View3D_SetTitle_button_function = lambda:View3D_SetTitle_button_function(self.guiWindow)
        self.guiWindow.View3D_plot_button_function = lambda:View3D_plot_button_function(self.guiWindow)
        self.guiWindow.View3D_toggle_mode_function = lambda: View3D_toggle_mode_function(self.guiWindow)
        self.guiWindow.View3D_toggle_plotCurratAxe_function = lambda: View3D_toggle_plotCurratAxe_function(self.guiWindow)
        self.guiWindow.View3D_changed_CurratAxeList_function = lambda: View3D_changed_CurratAxeList_function(self.guiWindow)
        self.guiWindow.View3D_DataSet_selectionChanged_function = lambda: View3D_DataSet_selectionChanged_function(self.guiWindow)
        self.guiWindow.View3D_Grid_checkBox_toggled_function = lambda: View3D_Grid_checkBox_toggled_function(self.guiWindow)
        self.guiWindow.View3D_indexChanged = lambda index: indexChanged(self.guiWindow,index)
        self.guiWindow.View3D_saveToCSV = lambda: saveToCSV(self.guiWindow)

        for key,value in self.__dict__.items():
            if 'View3D' in key:
                self.guiWindow.ui.__dict__[key] = value

        ## Set up figure manager. Is to be initialized between manager initialization but before their setup
        self.guiWindow.figureListView3D = MatplotlibFigureList(combobox=self.View3D_figureList_comboBox)
        self.guiWindow.figureList.append(self.guiWindow.figureListView3D)
        self.View3D_figureList_comboBox.setModel(self.guiWindow.figureListView3D)

        self.mplListDelegate = MatplotlibFigureListDelegate()
        

    def setup(self):
        self.guiWindow.ui.View3D_plot_button.clicked.connect(self.guiWindow.View3D_plot_button_function)
        self.guiWindow.ui.View3D_setCAxis_button.clicked.connect(self.guiWindow.View3D_setCAxis_button_function)
        self.guiWindow.ui.View3D_SetTitle_button.clicked.connect(self.guiWindow.View3D_SetTitle_button_function)

        self.guiWindow.ui.View3D_SetTitle_lineEdit.returnPressed.connect(self.titleEnterPressed)
        self.guiWindow.ui.View3D_CAxisMax_lineEdit.returnPressed.connect(self.CAxisChanged)
        self.guiWindow.ui.View3D_CAxisMin_lineEdit.returnPressed.connect(self.CAxisChanged)
        
        # Radiobutton to select viewing type
        self.guiWindow.ui.View3D_Mode_Viewer3D_radioButton.toggled.connect(self.guiWindow.View3D_toggle_mode_function)
        self.guiWindow.ui.View3D_CurratAxe_checkBox.toggled.connect(self.guiWindow.View3D_toggle_plotCurratAxe_function)
        self.guiWindow.ui.View3D_Grid_checkBox.toggled.connect(self.guiWindow.View3D_Grid_checkBox_toggled_function)

        self.guiWindow.DataSetSelectionModel.selectionChanged.connect(self.guiWindow.View3D_DataSet_selectionChanged_function)
        self.guiWindow.DataSetModel.dataChanged.connect(self.guiWindow.View3D_DataSet_selectionChanged_function)
        self.guiWindow.figureListView3D.view.currentIndexChanged.connect(self.guiWindow.View3D_indexChanged)
        self.guiWindow.ui.View3D_figureList_comboBox.setItemDelegate(self.mplListDelegate)

        self.guiWindow.ui.View3D_figureToCSV_pushButton.clicked.connect(self.guiWindow.View3D_saveToCSV)

    def titleEnterPressed(self):
        if self.guiWindow.ui.View3D_SetTitle_button.isEnabled():
            self.guiWindow.View3D_SetTitle_button_function()

    def CAxisChanged(self):
        if self.guiWindow.ui.View3D_setCAxis_button.isEnabled():
            self.guiWindow.View3D_setCAxis_button_function()

    def curratAxeList(self):
        if hasattr(self,'BraggListWindow'): # If a window is open, use it
            self.braggPoints = self.BraggListWindow.getData()
        else:
            self.BraggListWindow = BraggListManager.BraggListManager(BraggList=self.braggPoints)
            self.guiWindow.windows.append(self.BraggListWindow)
            self.BraggListWindow.show()


import sys,os
sys.path.append('..')

try:
    from MJOLNIRGui.src.main.python._tools import loadUI, dialog
except ImportError:
    from _tools import loadUI, dialog
from os import path
import numpy as np
from PyQt5 import QtWidgets, uic, QtCore, QtGui
from MJOLNIR.Data import DataFile


class customCheckBox(QtWidgets.QCheckBox):
    def __init__(self,*args,name=None,manager=None,**kwargs):
        super().__init__(*args,**kwargs)
        self.stateChanged.connect(lambda state: manager.boxTicked(self,state))
        self.name = name


ElectronicLogBookManagerBase, ElectronicLogBookManagerForm = loadUI('ElectronicLogBookWidget.ui')

class ElectronicLogBookManager(ElectronicLogBookManagerBase, ElectronicLogBookManagerForm):
    def __init__(self, parent=None, guiWindow=None):
        super(ElectronicLogBookManager, self).__init__(parent)
        self.setupUi(self)
        self.guiWindow = guiWindow
        self.setWindowIcon(QtGui.QIcon(self.guiWindow.AppContext.get_resource('Icons/Own/book--pencil.png')))

        if not hasattr(self.guiWindow,'logbookPreset'):
            self.guiWindow.logbookPreset = []

        self.possibleParameters = DataFile.possibleAttributes
        self.initElectronicLogBookManager()
        self.initialMove()

        if 'saveToFile' in self.guiWindow.logbookPreset:
            self.ElectronicLogBook_savetofile_radioButton.setChecked(True)
        else:
            self.ElectronicLogBook_printinlog_radioButton.setChecked(True)

        self.checkDataSet()

    def initElectronicLogBookManager(self):
        self.layout = QtWidgets.QVBoxLayout()
        self.dummyWidget = QtWidgets.QWidget()

        for par in self.possibleParameters:
            horizontal = QtWidgets.QHBoxLayout()
            label = QtWidgets.QLabel(par)
            tickBox = customCheckBox(name=par,manager=self)
            tickBox.setObjectName(par+'_checkBox')

            if not par in self.guiWindow.logbookPreset:
                checked = False
            else:
                checked = 2
            tickBox.setCheckState(checked)
            setattr(self,par+'_checkBox',tickBox)
            horizontal.addWidget(label)
            horizontal.addWidget(tickBox)
            self.layout.addLayout(horizontal)

        self.dummyWidget.setLayout(self.layout)
        self.parametersLayout.setWidget(self.dummyWidget)
        self.ElectronicLogBook_load_button.clicked.connect(self.load)
        self.ElectronicLogBook_dataset_button.clicked.connect(self.loadFromDataSet)
        self.ElectronicLogBook_savetofile_radioButton.toggled.connect(self.saveToFileChanged)
        self.guiWindow.DataSetModel.layoutChanged.connect(self.checkDataSet)
        
    
    def getPars(self):
        pars = self.guiWindow.logbookPreset
        for aux in ['saveToFile','printToLog']:
            if aux in pars: # remove the auxilliary entries
                del pars[pars.index(aux)]
        if not 'name' in pars:
            pars.insert(0,'name')
        return pars

    def load(self):
        pars = self.getPars()
        folder = self.guiWindow.getCurrentDirectory()
        allowedRawFilesString = 'Raw ('+' '.join(['*.'+str(x) for x in DataFile.supportedRawFormats])+')'
        allowedAllFilesString = 'All Files (*)'
        allowedString = ';;'.join([allowedRawFilesString,allowedAllFilesString])
        files, _ = QtWidgets.QFileDialog.getOpenFileNames(self,"Open data Files", folder,allowedString)
        if files is None:
            return False
        if len(files) == 0:
            return False
        
        result = DataFile.shallowRead(files,pars)
        self.returnResult(pars,result)

    def loadFromDataSet(self):
        pars = self.getPars()
        ds = self.guiWindow.DataSetModel.getCurrentDataSet()
        files = [df.fileLocation for df in ds.dataFiles]
        if files is None:
            self.guiWindow.writeToStatus('Selected DataSet is empty')
            dialog('Selected DataSet is empty')
            return False
        if len(files) == 0:
            self.guiWindow.writeToStatus('Selected DataSet is empty')
            dialog('Selected DataSet is empty')
            return False
        
        result = DataFile.shallowRead(files,pars)
        self.returnResult(pars,result)
        
    def beautify(self,pars,result):
        returnString = []
        for fileResult in result:
            localReturnString = []
            for par,res in zip(pars,fileResult):
                if hasattr(res,'shape'):
                    if res.size == 1:
                        try:
                            localReturnString.append(' '+par+': '+str(res[0]))
                        except IndexError: # res = np.array(None) and the above throws an error!
                            localReturnString.append(' '+par+': None')
                    else:
                        try:
                            m = np.min(res)
                            M = np.max(res)
                        except TypeError:
                            localReturnString.append(' '+par+': '+str(res[0]))
                        else:
                            if np.isclose(m,M):
                                localReturnString.append(' '+par+': '+str(res[0]))
                            else:
                                localReturnString.append(' '+par+': [{} - {}]'.format(str(m),str(M)))
                else:
                    localReturnString.append(' '+par+': '+str(res))
            returnString.append(' '.join(localReturnString))

        returnString = '\n'.join([x for x in returnString])

        return returnString

    def returnResult(self,pars,result): # function to return result to user (either file or log)
        result = self.beautify(pars,result)
        if self.ElectronicLogBook_savetofile_radioButton.isChecked():
            saveLocation,_ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File',None,"Text (*.txt)")
            if len(saveLocation) == 0:
                return False
            with open(saveLocation,'w') as f:
                f.writelines(result)
        else:
            self.guiWindow.writeToStatus(result)
        # After successful return, close window
        self.close()

    def checkDataSet(self):
        if self.guiWindow.DataSetModel.getCurrentDatasetIndex() is None:
            self.ElectronicLogBook_dataset_button.setDisabled(True)
        else:
            self.ElectronicLogBook_dataset_button.setDisabled(False)

    def initialMove(self):
        qtRectangle = self.frameGeometry()
        centerPoint = QtCore.QPoint(qtRectangle.center().x(),int(qtRectangle.height()/2.0)+5) # Move widget 5 pixel from top edge
        
        qtRectangle.moveCenter(centerPoint)
        self.move(qtRectangle.topLeft())

    def boxTicked(self,box,state):
        if state: # Going to true
            if not box.name in self.guiWindow.logbookPreset:
                self.guiWindow.logbookPreset.append(box.name)
        else:
            if box.name in self.guiWindow.logbookPreset:
                idx = self.guiWindow.logbookPreset.index(box.name)
                del self.guiWindow.logbookPreset[idx]

    def saveToFileChanged(self,state):
        if state: # changed to true
            if 'printToLog' in self.guiWindow.logbookPreset:
                del self.guiWindow.logbookPreset[self.guiWindow.logbookPreset.index('printToLog')]
            self.guiWindow.logbookPreset.append('saveToFile')
        else:
            if 'saveToFile' in self.guiWindow.logbookPreset:
                del self.guiWindow.logbookPreset[self.guiWindow.logbookPreset.index('saveToFile')]
            self.guiWindow.logbookPreset.append('printToLog')
        
        
        


import sys
try:
    from StateMachine import State,AllFalseBut,AllTrue,AllFalse,AllTrueBut
except ImportError:
    from .StateMachine import State,AllFalseBut,AllTrue,AllFalse,AllTrueBut



import numpy as np

import platform

platform.system()

if platform.system() == 'Windows':
    markerColor = 'lightblue'
else:
    markerColor = 'palette(Link)'

highlightStyle = "background-color: palette(Link); color: white"
normalStyle = "background-color: palette(midlight)"

## enterStateFunction
def AllEnabled(StateMachine):
    names = [item.objectName() for item in StateMachine.guiWindow.blockItems]
    enabled = np.ones(len(names))
    #enabledNames = ['DataSet_NewDataSet_button',]
    for item,enable in zip(StateMachine.guiWindow.blockItems,enabled):
        item.setEnabled(int(enable))

    blue = AllFalseBut(['View3D_plot_button','QELine_plot_button','QPlane_plot_button'],enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,blue):
        if enable:
            item.setStyleSheet(highlightStyle)
        else:
            item.setStyleSheet(normalStyle)

def ConvertedEnabled(StateMachine):
    names = [item.objectName() for item in StateMachine.guiWindow.blockItems]
    enabled = np.ones(len(names))
    enabledNames = ['Raw1D_fit_button']
    
    enabled = AllTrueBut(enabledNames,enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,enabled):
        item.setEnabled(int(enable))

    blue = AllFalseBut(['View3D_plot_button','QELine_plot_button','QPlane_plot_button','Cut1D_plot_button','Cut1D_Generate1D_button'],enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,blue):
        if enable:
            item.setStyleSheet(highlightStyle)
        else:
            item.setStyleSheet(normalStyle)

def RawEnabled(StateMachine):
    names = [item.objectName() for item in StateMachine.guiWindow.blockItems]
    enabled = np.ones(len(names))
    enabledNames = ['DataSet_NewDataSet_button','DataSet_DeleteDataSet_button',
    'DataSet_AddFiles_button','DataSet_DeleteFiles_button','DataSet_convertData_button','DataSet_binning_comboBox',
    'View3D_plot_button','QELine_plot_button','QPlane_plot_button','Raw1D_plot_button','Cut1D_plot_button','View3D_CurratAxe_button']
    
    enabled = AllFalseBut(enabledNames,enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,enabled):
        item.setEnabled(int(enable))

    blue = AllFalseBut(['DataSet_convertData_button','Raw1D_plot_button'],enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,blue):
        if enable:
            item.setStyleSheet(highlightStyle)
        else:
            item.setStyleSheet(normalStyle)


def PartialEnabled(StateMachine):
    names = [item.objectName() for item in StateMachine.guiWindow.blockItems]
    enabled = np.ones(len(names))
    enabledNames = ['DataSet_NewDataSet_button','DataSet_DeleteDataSet_button',
    'DataSet_AddFiles_button',
    'View3D_plot_button','QELine_plot_button','QPlane_plot_button','Raw1D_plot_button','Cut1D_plot_button','View3D_CurratAxe_button']
    
    enabled = AllFalseBut(enabledNames,enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,enabled):
        item.setEnabled(int(enable))

    blue = AllFalseBut(['DataSet_AddFiles_button'],enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,blue):
        if enable:
            item.setStyleSheet(highlightStyle)
        else:
            item.setStyleSheet(normalStyle)
    


def EmptyEnabled(StateMachine):
    names = [item.objectName() for item in StateMachine.guiWindow.blockItems]
    enabled = np.ones(len(names))
    enabledNames = ['DataSet_NewDataSet_button','View3D_plot_button','QELine_plot_button','QPlane_plot_button','Raw1D_plot_button','Cut1D_plot_button','View3D_CurratAxe_button']
    
    enabled = AllFalseBut(enabledNames,enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,enabled):
        item.setEnabled(int(enable))

    blue = AllFalseBut(['DataSet_NewDataSet_button'],enabled,names)
    for item,enable in zip(StateMachine.guiWindow.blockItems,blue):
        if enable:
            item.setStyleSheet(highlightStyle)
        else:
            item.setStyleSheet(normalStyle)
    if hasattr(StateMachine.guiWindow.ui,'View3D_SetTitle_lineEdit'):
        StateMachine.guiWindow.ui.View3D_SetTitle_lineEdit.setPlaceholderText('')
    

## Transition functions

def transitionEmptyPartial(StateMachine): # What is required to transition from empty to partial
    return StateMachine.guiWindow.DataSetModel.rowCount(None)>0

def transitionPartialRaw(StateMachine): # What is required to transition from partial to raw
    if StateMachine.guiWindow.DataSetModel.getCurrentDataSet() is None:
        return False
    count = len(StateMachine.guiWindow.DataSetModel.getCurrentDataSet())
    if not count is None:
        return count>0
    else:
        return False

def transitionRawConverted(StateMachine): # What is required to transition from partial to raw
    ds = StateMachine.guiWindow.DataSetModel.getCurrentDataSet()
    
    if not ds is None:
        val = len(ds.convertedFiles)>0
    else:
        val = False
    
    return val
### Functions to force transition

def forceTransitionEmptyPartial(StateMachine): # add DataSet
    StateMachine.guiWindow.DataSet_NewDataSet_button_function()
    return transitionEmptyPartial(StateMachine)

def forceTransitionPartialRaw(StateMachine): # add DataFile
    StateMachine.guiWindow.DataSet_AddFiles_button_function()
    return transitionPartialRaw(StateMachine)

def forceTansitionRawConverted(StateMachine): # add DataFile
    StateMachine.guiWindow.convert()
    return True#transitionRawConverted(StateMachine)


### States for state machine

# Allows plotting, fitting and all the fun
converted = State('Converted',nextState=None,enterStateFunction=ConvertedEnabled) 

# Has DataSet, DataFiles but not converded yet
raw = State('Raw',enterStateFunction=RawEnabled,transitionRequirement=transitionRawConverted,
                transitionFunction=forceTansitionRawConverted,nextState=converted)

# Has DataSet but no DataFiles
partial = State('Partial',enterStateFunction=PartialEnabled,transitionRequirement=transitionPartialRaw,
                transitionFunction=forceTransitionPartialRaw,nextState=raw)

# Has no DataSet
empty = State('Empty',enterStateFunction=EmptyEnabled,transitionRequirement=transitionEmptyPartial,
                transitionFunction=forceTransitionEmptyPartial,nextState=partial)



#from fbs_runtime.application_context.PyQt5 import ApplicationContext
import sys
try:
    from MJOLNIR_GUI import MJOLNIRMainWindow,updateSplash
except ImportError:
    import os
    os.chdir(os.path.dirname(__file__))
    from MJOLNIRGui.src.main.python.MJOLNIR_GUI import MJOLNIRMainWindow,updateSplash
    
from PyQt5 import QtWidgets, QtGui, QtCore
import datetime


from fbs_runtime.application_context.PyQt5 import ApplicationContext, \
    cached_property
    

class AppContext(ApplicationContext):
    def __init__(self,*args,**kwargs):
        super().__init__(*args,**kwargs)
        self.splash = QtWidgets.QSplashScreen(QtGui.QPixmap(self.get_resource('splash.png')))                                    
        
        self.splash.show()
        
        self.timer = QtCore.QTimer() 
        
        updateInterval = 400 # ms
        originalTime = datetime.datetime.now()
        
        updater = lambda:updateSplash(self.splash,originalTime=originalTime,updateInterval=updateInterval)
        updater()
        
        
        self.timer.timeout.connect(updater) 
        self.timer.setInterval(updateInterval)
        self.timer.start()
        QtWidgets.QApplication.processEvents()

        

    def run(self):
        
        QtWidgets.QApplication.processEvents()
        self.splash.finish(self.main_window)
        self.main_window.show()

        if len(sys.argv)==2:
            self.main_window.loadGui(presetFileLocation=sys.argv[1])
        return self.app.exec_()

    @cached_property
    def main_window(self):
        QtWidgets.QApplication.processEvents()
        res = MJOLNIRMainWindow(self)
        self.timer.stop()
        return res # Pass context to the window.


def main():
    appctxt = AppContext()
    exit_code = appctxt.run()
    sys.exit(exit_code)

if __name__ == '__main__':
    import os
    os.environ["QT_LOGGING_RULES"] = "*.debug=false"
    main()
